# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for click_package.osextras."""

from __future__ import print_function
__all__ = [
    'TestOSExtrasNative',
    'TestOSExtrasPython',
    ]


import os

from gi.repository import Click, GLib

from click_package import osextras
from click_package.tests.helpers import TestCase, mock, touch


class TestOSExtrasBaseMixin:
    def test_ensuredir_previously_missing(self):
        new_dir = os.path.join(self.temp_dir, "dir")
        self.mod.ensuredir(new_dir)
        self.assertTrue(os.path.isdir(new_dir))

    def test_ensuredir_previously_present(self):
        new_dir = os.path.join(self.temp_dir, "dir")
        os.mkdir(new_dir)
        self.mod.ensuredir(new_dir)
        self.assertTrue(os.path.isdir(new_dir))

    def test_find_on_path_missing_environment(self):
        os.environ.pop("PATH", None)
        self.assertFalse(self.mod.find_on_path("ls"))

    def test_find_on_path_present_executable(self):
        bin_dir = os.path.join(self.temp_dir, "bin")
        program = os.path.join(bin_dir, "program")
        touch(program)
        os.chmod(program, 0o755)
        os.environ["PATH"] = bin_dir
        self.assertTrue(self.mod.find_on_path("program"))

    def test_find_on_path_present_not_executable(self):
        bin_dir = os.path.join(self.temp_dir, "bin")
        touch(os.path.join(bin_dir, "program"))
        os.environ["PATH"] = bin_dir
        self.assertFalse(self.mod.find_on_path("program"))

    def test_find_on_path_requires_regular_file(self):
        bin_dir = os.path.join(self.temp_dir, "bin")
        self.mod.ensuredir(os.path.join(bin_dir, "subdir"))
        os.environ["PATH"] = bin_dir
        self.assertFalse(self.mod.find_on_path("subdir"))

    def test_unlink_file_present(self):
        path = os.path.join(self.temp_dir, "file")
        touch(path)
        self.mod.unlink_force(path)
        self.assertFalse(os.path.exists(path))

    def test_unlink_file_missing(self):
        path = os.path.join(self.temp_dir, "file")
        self.mod.unlink_force(path)
        self.assertFalse(os.path.exists(path))

    def test_symlink_file_present(self):
        path = os.path.join(self.temp_dir, "link")
        touch(path)
        self.mod.symlink_force("source", path)
        self.assertTrue(os.path.islink(path))
        self.assertEqual("source", os.readlink(path))

    def test_symlink_link_present(self):
        path = os.path.join(self.temp_dir, "link")
        os.symlink("old", path)
        self.mod.symlink_force("source", path)
        self.assertTrue(os.path.islink(path))
        self.assertEqual("source", os.readlink(path))

    def test_symlink_missing(self):
        path = os.path.join(self.temp_dir, "link")
        self.mod.symlink_force("source", path)
        self.assertTrue(os.path.islink(path))
        self.assertEqual("source", os.readlink(path))

    def test_umask(self):
        old_mask = os.umask(0o040)
        try:
            self.assertEqual(0o040, self.mod.get_umask())
            os.umask(0o002)
            self.assertEqual(0o002, self.mod.get_umask())
        finally:
            os.umask(old_mask)


class TestOSExtrasNative(TestCase, TestOSExtrasBaseMixin):
    def setUp(self):
        super(TestOSExtrasNative, self).setUp()
        self.use_temp_dir()
        self.mod = Click

    def test_ensuredir_error(self):
        path = os.path.join(self.temp_dir, "file")
        touch(path)
        self.assertRaisesFileError(mock.ANY, self.mod.ensuredir, path)

    def test_dir_read_name_directory_present(self):
        new_dir = os.path.join(self.temp_dir, "dir")
        touch(os.path.join(new_dir, "file"))
        d = Click.Dir.open(new_dir, 0)
        self.assertEqual("file", d.read_name())
        self.assertIsNone(d.read_name())

    def test_dir_read_name_directory_missing(self):
        new_dir = os.path.join(self.temp_dir, "dir")
        d = Click.Dir.open(new_dir, 0)
        self.assertIsNone(d.read_name())

    def test_dir_open_error(self):
        not_dir = os.path.join(self.temp_dir, "file")
        touch(not_dir)
        self.assertRaisesFileError(
            GLib.FileError.NOTDIR, Click.Dir.open, not_dir, 0)

    def test_unlink_error(self):
        path = os.path.join(self.temp_dir, "dir")
        os.mkdir(path)
        self.assertRaisesFileError(mock.ANY, self.mod.unlink_force, path)

    def test_symlink_unlink_error(self):
        path = os.path.join(self.temp_dir, "dir")
        os.mkdir(path)
        self.assertRaisesFileError(
            mock.ANY, self.mod.symlink_force, "source", path)

    def test_symlink_error(self):
        path = os.path.join(self.temp_dir, "dir", "file")
        self.assertRaisesFileError(
            mock.ANY, self.mod.symlink_force, "source", path)


class TestOSExtrasPython(TestCase, TestOSExtrasBaseMixin):
    def setUp(self):
        super(TestOSExtrasPython, self).setUp()
        self.use_temp_dir()
        self.mod = osextras

    def test_ensuredir_oserror(self):
        path = os.path.join(self.temp_dir, "file")
        touch(path)
        self.assertRaises(OSError, self.mod.ensuredir, path)

    def test_listdir_directory_present(self):
        new_dir = os.path.join(self.temp_dir, "dir")
        touch(os.path.join(new_dir, "file"))
        self.assertEqual(["file"], osextras.listdir_force(new_dir))

    def test_listdir_directory_missing(self):
        new_dir = os.path.join(self.temp_dir, "dir")
        self.assertEqual([], osextras.listdir_force(new_dir))

    def test_listdir_oserror(self):
        not_dir = os.path.join(self.temp_dir, "file")
        touch(not_dir)
        self.assertRaises(OSError, osextras.listdir_force, not_dir)

    def test_unlink_oserror(self):
        path = os.path.join(self.temp_dir, "dir")
        os.mkdir(path)
        self.assertRaises(OSError, self.mod.unlink_force, path)

    def test_symlink_unlink_oserror(self):
        path = os.path.join(self.temp_dir, "dir")
        os.mkdir(path)
        self.assertRaises(OSError, self.mod.symlink_force, "source", path)

    def test_symlink_oserror(self):
        path = os.path.join(self.temp_dir, "dir", "file")
        self.assertRaises(OSError, self.mod.symlink_force, "source", path)



# Copyright (C) 2013 Canonical Ltd.
# Authors: Colin Watson <cjwatson@ubuntu.com>,
#          Brian Murray <brian@ubuntu.com>
#          Michael Vogt <mvo@ubuntu.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Chroot management for building Click packages."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    "ClickChroot",
    "ClickChrootException",
    "ClickChrootAlreadyExistsException",
    "ClickChrootDoesNotExistException",
    ]

try:
    from urllib.error import URLError
    from urllib.request import urlopen
except ImportError:
    from urllib2 import URLError, urlopen
import os
import pwd
import fileinput
import re
import shutil
import stat
import subprocess
import sys
from textwrap import dedent
from xml.etree import ElementTree


framework_base = {
    "ubuntu-sdk-13.10": "ubuntu-sdk-13.10",
    # 14.04
    "ubuntu-sdk-14.04-html": "ubuntu-sdk-14.04",
    "ubuntu-sdk-14.04-papi": "ubuntu-sdk-14.04",
    "ubuntu-sdk-14.04-qml": "ubuntu-sdk-14.04",
    # 14.10
    "ubuntu-sdk-14.10-html": "ubuntu-sdk-14.10",
    "ubuntu-sdk-14.10-papi": "ubuntu-sdk-14.10",
    "ubuntu-sdk-14.10-qml": "ubuntu-sdk-14.10",
    # 15.04
    "ubuntu-sdk-15.04-html": "ubuntu-sdk-15.04",
    "ubuntu-sdk-15.04-papi": "ubuntu-sdk-15.04",
    "ubuntu-sdk-15.04-qml": "ubuntu-sdk-15.04",
    # 15.10
    "ubuntu-sdk-15.10-html-dev1": "ubuntu-sdk-15.10-dev1",
    "ubuntu-sdk-15.10-papi-dev1": "ubuntu-sdk-15.10-dev1",
    "ubuntu-sdk-15.10-qml-dev1": "ubuntu-sdk-15.10-dev1",
    }


framework_series = {
    "ubuntu-sdk-13.10": "saucy",
    "ubuntu-sdk-14.04": "trusty",
    "ubuntu-sdk-14.10": "utopic",
    "ubuntu-sdk-15.04": "vivid",
    "ubuntu-sdk-15.10": "wily",
    }


# Please keep the lists of package names sorted.
extra_packages = {
    "ubuntu-sdk-13.10": [
        "libqt5opengl5-dev:{TARGET}",
        "libqt5svg5-dev:{TARGET}",
        "libqt5v8-5-dev:{TARGET}",
        "libqt5webkit5-dev:{TARGET}",
        "libqt5xmlpatterns5-dev:{TARGET}",
        "qmlscene:{TARGET}",
        "qt3d5-dev:{TARGET}",
        "qt5-default:{TARGET}",
        "qt5-qmake:{TARGET}",
        "qtbase5-dev:{TARGET}",
        "qtdeclarative5-dev:{TARGET}",
        "qtmultimedia5-dev:{TARGET}",
        "qtquick1-5-dev:{TARGET}",
        "qtscript5-dev:{TARGET}",
        "qtsensors5-dev:{TARGET}",
        "qttools5-dev:{TARGET}",
        "ubuntu-ui-toolkit-doc",
        ],
    "ubuntu-sdk-14.04": [
        "cmake",
        "google-mock:{TARGET}",
        "intltool",
        "libboost1.54-dev:{TARGET}",
        "libjsoncpp-dev:{TARGET}",
        "libprocess-cpp-dev:{TARGET}",
        "libproperties-cpp-dev:{TARGET}",
        "libqt5svg5-dev:{TARGET}",
        "libqt5webkit5-dev:{TARGET}",
        "libqt5xmlpatterns5-dev:{TARGET}",
        "libunity-scopes-dev:{TARGET}",
        # bug #1316930, needed for autopilot
        "python3",
        "qmlscene:{TARGET}",
        "qt3d5-dev:{TARGET}",
        "qt5-default:{TARGET}",
        "qtbase5-dev:{TARGET}",
        "qtdeclarative5-dev:{TARGET}",
        "qtdeclarative5-dev-tools",
        "qtlocation5-dev:{TARGET}",
        "qtmultimedia5-dev:{TARGET}",
        "qtscript5-dev:{TARGET}",
        "qtsensors5-dev:{TARGET}",
        "qttools5-dev:{TARGET}",
        "qttools5-dev-tools:{TARGET}",
        "ubuntu-ui-toolkit-doc",
        ],
    "ubuntu-sdk-14.10": [
        "cmake",
        "cmake-extras",
        "google-mock:{TARGET}",
        "intltool",
        "libboost1.55-dev:{TARGET}",
        "libcontent-hub-dev:{TARGET}",
        "libjsoncpp-dev:{TARGET}",
        "libnet-cpp-dev:{TARGET}",
        "libprocess-cpp-dev:{TARGET}",
        "libproperties-cpp-dev:{TARGET}",
        "libqt5keychain0:{TARGET}",
        "libqt5sensors5-dev:{TARGET}",
        "libqt5svg5-dev:{TARGET}",
        "libqt5webkit5-dev:{TARGET}",
        "libqt5xmlpatterns5-dev:{TARGET}",
        "libunity-scopes-dev:{TARGET}",
        # bug #1316930, needed for autopilot
        "python3",
        "qml-module-qt-labs-settings:{TARGET}",
        "qml-module-qtmultimedia:{TARGET}",
        "qml-module-qtquick-layouts:{TARGET}",
        "qml-module-qtsensors:{TARGET}",
        "qml-module-qtwebkit:{TARGET}",
        "qmlscene:{TARGET}",
        "qt3d5-dev:{TARGET}",
        "qt5-default:{TARGET}",
        "qtdeclarative5-accounts-plugin:{TARGET}",
        "qtdeclarative5-dev-tools",
        "qtdeclarative5-folderlistmodel-plugin:{TARGET}",
        "qtdeclarative5-localstorage-plugin:{TARGET}",
        "qtdeclarative5-online-accounts-client0.1:{TARGET}",
        "qtdeclarative5-particles-plugin:{TARGET}",
        "qtdeclarative5-poppler1.0:{TARGET}",
        "qtdeclarative5-qtlocation-plugin:{TARGET}",
        "qtdeclarative5-qtorganizer-plugin:{TARGET}",
        "qtdeclarative5-qtpositioning-plugin:{TARGET}",
        "qtdeclarative5-u1db1.0:{TARGET}",
        "qtdeclarative5-ubuntu-content0.1:{TARGET}",
        "qtdeclarative5-ubuntu-download-manager0.1:{TARGET}",
        "qtdeclarative5-ubuntu-mediascanner0.1:{TARGET}",
        "qtdeclarative5-ubuntu-syncmonitor0.1:{TARGET}",
        "qtdeclarative5-ubuntu-telephony-phonenumber0.1:{TARGET}",
        "qtdeclarative5-ubuntu-ui-toolkit-plugin:{TARGET}",
        "qtdeclarative5-usermetrics0.1:{TARGET}",
        "qtdeclarative5-xmllistmodel-plugin:{TARGET}",
        "qtlocation5-dev:{TARGET}",
        "qtmultimedia5-dev:{TARGET}",
        "qtscript5-dev:{TARGET}",
        "qttools5-dev:{TARGET}",
        "qttools5-dev-tools:{TARGET}",
        "ubuntu-html5-theme:{TARGET}",
        "ubuntu-ui-toolkit-doc",
        ],
    "ubuntu-sdk-15.04": [
        # the sdk libs
        "ubuntu-sdk-libs:{TARGET}",
        "ubuntu-sdk-libs-dev:{TARGET}",
        # the native build tools
        "ubuntu-sdk-libs-tools",
        # FIXME: see
        #  http://pad.lv/~mvo/oxide/crossbuild-friendly/+merge/234093
        # we help the apt resolver here until the
        #  oxideqt-codecs/oxidec-codecs-extras is sorted
        "oxideqt-codecs-extra",
        ],
    "ubuntu-sdk-15.10-dev1": [
        # the sdk libs
        "ubuntu-sdk-libs:{TARGET}",
        "ubuntu-sdk-libs-dev:{TARGET}",
        # the native build tools
        "ubuntu-sdk-libs-tools",
        # FIXME: see
        #  http://pad.lv/~mvo/oxide/crossbuild-friendly/+merge/234093
        # we help the apt resolver here until the
        #  oxideqt-codecs/oxidec-codecs-extras is sorted
        "oxideqt-codecs-extra",
        ],
    }


primary_arches = ["amd64", "i386"]


non_meta_re = re.compile(r'^[a-zA-Z0-9+,./:=@_-]+$')


GEOIP_SERVER = "http://geoip.ubuntu.com/lookup"

overlay_ppa = "ci-train-ppa-service/stable-phone-overlay"


def get_geoip_country_code_prefix():
    click_no_local_mirror = os.environ.get('CLICK_NO_LOCAL_MIRROR', 'auto')
    if click_no_local_mirror == '1':
        return ""
    try:
        with urlopen(GEOIP_SERVER) as f:
            xml_data = f.read()
        et = ElementTree.fromstring(xml_data)
        cc = et.find("CountryCode")
        if not cc:
            return ""
        return cc.text.lower()+"."
    except (ElementTree.ParseError, URLError):
        pass
    return ""


def generate_sources(series, native_arch, target_arch,
                     archive_mirror, ports_mirror, components):
    """Generate a list of strings for apts sources.list.
    Arguments:
    series -- the distro series (e.g. vivid)
    native_arch -- the native architecture (e.g. amd64)
    target_arch -- the target architecture (e.g. armhf)
    archive_mirror -- main mirror, e.g. http://archive.ubuntu.com/ubuntu
    ports_mirror -- ports mirror, e.g. http://ports.ubuntu.com/ubuntu-ports
    components -- the components as string, e.g. "main restricted universe"
    """
    pockets = ['%s' % series]
    for pocket in ['updates', 'security']:
        pockets.append('%s-%s' % (series, pocket))
    sources = []
    # write binary lines
    arches = [target_arch]
    if native_arch != target_arch:
        arches.append(native_arch)
    for arch in arches:
        if arch not in primary_arches:
            mirror = ports_mirror
        else:
            mirror = archive_mirror
        for pocket in pockets:
            sources.append("deb [arch=%s] %s %s %s" %
                           (arch, mirror, pocket, components))
    # write source lines
    for pocket in pockets:
        sources.append("deb-src %s %s %s" %
                       (archive_mirror, pocket, components))
    return sources


def shell_escape(command):
    escaped = []
    for arg in command:
        if non_meta_re.match(arg):
            escaped.append(arg)
        else:
            escaped.append("'%s'" % arg.replace("'", "'\\''"))
    return " ".join(escaped)


def strip_dev_series_from_framework(framework):
    """Remove trailing -dev[0-9]+ from a framework name"""
    return re.sub(r'^(.*)-dev[0-9]+$', r'\1', framework)


class ClickChrootException(Exception):
    """A generic issue with the chroot"""
    pass


class ClickChrootAlreadyExistsException(ClickChrootException):
    """The chroot already exists"""
    pass


class ClickChrootDoesNotExistException(ClickChrootException):
    """A chroot with that name does not exist yet"""
    pass


class ClickChroot:

    DAEMON_POLICY = dedent("""\
    #!/bin/sh
    while true; do
        case "$1" in
          -*) shift ;;
          makedev) exit 0;;
          x11-common) exit 0;;
          *) exit 101;;
        esac
    done
    """)

    def __init__(self, target_arch, framework, name=None, series=None,
                 session=None, chroots_dir=None):
        self.target_arch = target_arch
        self.framework = strip_dev_series_from_framework(framework)
        if name is None:
            name = "click"
        self.name = name
        if series is None:
            series = framework_series[self.framework_base]
        self.series = series
        self.session = session
        system_arch = subprocess.check_output(
            ["dpkg", "--print-architecture"],
            universal_newlines=True).strip()
        self.native_arch = self._get_native_arch(system_arch, self.target_arch)
        if chroots_dir is None:
            chroots_dir = "/var/lib/schroot/chroots"
        self.chroots_dir = chroots_dir

        if "SUDO_USER" in os.environ:
            self.user = os.environ["SUDO_USER"]
        elif "PKEXEC_UID" in os.environ:
            self.user = pwd.getpwuid(int(os.environ["PKEXEC_UID"])).pw_name
        else:
            self.user = pwd.getpwuid(os.getuid()).pw_name
        self.dpkg_architecture = self._dpkg_architecture()

    def _get_native_arch(self, system_arch, target_arch):
        """Determine the proper native architecture for a chroot.

        Some combinations of system and target architecture do not require
        cross-building, so in these cases we just create a chroot suitable
        for native building.
        """
        if (system_arch, target_arch) in (
                ("amd64", "i386"),
                # This will only work if the system is running a 64-bit
                # kernel; but there's no alternative since no i386-to-amd64
                # cross-compiler is available in the Ubuntu archive.
                ("i386", "amd64"),
                ):
            return target_arch
        else:
            return system_arch

    def _dpkg_architecture(self):
        dpkg_architecture = {}
        command = ["dpkg-architecture", "-a%s" % self.target_arch]
        env = dict(os.environ)
        env["CC"] = "true"
        # Force dpkg-architecture to recalculate everything rather than
        # picking up values from the environment, which will be present when
        # running the test suite under dpkg-buildpackage.
        for key in list(env):
            if key.startswith("DEB_BUILD_") or key.startswith("DEB_HOST_"):
                del env[key]
        lines = subprocess.check_output(
            command, env=env, universal_newlines=True).splitlines()
        for line in lines:
            try:
                key, value = line.split("=", 1)
            except ValueError:
                continue
            dpkg_architecture[key] = value
        if self.native_arch == self.target_arch:
            # We may have overridden the native architecture (see
            # _get_native_arch above), so we need to force DEB_BUILD_* to
            # match.
            for key in list(dpkg_architecture):
                if key.startswith("DEB_HOST_"):
                    new_key = "DEB_BUILD_" + key[len("DEB_HOST_"):]
                    dpkg_architecture[new_key] = dpkg_architecture[key]
        return dpkg_architecture

    def _get_overlayfs_name(self):
        for line in fileinput.input("/proc/filesystems"):
            if line.strip() == "nodev\toverlay":
                fileinput.close()
                return "overlay"
        return "overlayfs"

    def _generate_chroot_config(self, mount):
        admin_user = "root"
        users = []
        for key in ("users", "root-users", "source-root-users"):
            users.append("%s=%s,%s" % (key, admin_user, self.user))
        with open(self.chroot_config, "w") as target:
            target.write(dedent("""\
            [{full_name}]
            description=Build chroot for click packages on {target_arch}
            {users}
            type=directory
            profile=default
            setup.fstab=click/fstab
            # Not protocols or services see
            # debian bug 557730
            setup.nssdatabases=sbuild/nssdatabases
            union-type={overlayfs_name}
            directory={mount}
            """).format(full_name=self.full_name,
                        target_arch=self.target_arch,
                        users="\n".join(users),
                        mount=mount,
                        overlayfs_name=self._get_overlayfs_name()))

    def _generate_daemon_policy(self, mount):
        daemon_policy = "%s/usr/sbin/policy-rc.d" % mount
        with open(daemon_policy, "w") as policy:
            policy.write(self.DAEMON_POLICY)
        return daemon_policy

    def _generate_apt_proxy_file(self, mount, proxy):
        apt_conf_d = os.path.join(mount, "etc", "apt", "apt.conf.d")
        if not os.path.exists(apt_conf_d):
            os.makedirs(apt_conf_d)
        apt_conf_f = os.path.join(apt_conf_d, "99-click-chroot-proxy")
        if proxy:
            with open(apt_conf_f, "w") as f:
                f.write(dedent("""\
                // proxy settings copied by click chroot
                Acquire {
                    HTTP {
                        Proxy "%s";
                    };
                };
                """) % proxy)
        return apt_conf_f

    def _generate_finish_script(self, mount, build_pkgs):
        finish_script = "%s/finish.sh" % mount
        with open(finish_script, 'w') as finish:
            finish.write(dedent("""\
                #!/bin/bash
                set -e
                # Configure target arch
                dpkg --add-architecture {target_arch}
                # Reload package lists
                apt-get update || true
                # Pull down signature requirements
                apt-get -y --force-yes install gnupg ubuntu-keyring
            """).format(target_arch=self.target_arch))
            if self.series == "vivid":
                finish.write(dedent("""\
                    apt-get -y --force-yes install software-properties-common
                    add-apt-repository -y ppa:{ppa}
                    echo "Package: *"  \
                        > /etc/apt/preferences.d/stable-phone-overlay.pref
                    echo \
                        "Pin: release o=LP-PPA-{pin_ppa}" \
                        >> /etc/apt/preferences.d/stable-phone-overlay.pref
                    echo "Pin-Priority: 1001" \
                        >> /etc/apt/preferences.d/stable-phone-overlay.pref
                """).format(ppa=overlay_ppa,
                            pin_ppa=re.sub('/', '-', overlay_ppa)))
            finish.write(dedent("""\
                # Reload package lists
                apt-get update || true
                # Disable debconf questions
                # so that automated builds won't prompt
                echo set debconf/frontend Noninteractive | debconf-communicate
                echo set debconf/priority critical | debconf-communicate
                apt-get -y --force-yes dist-upgrade
                # Install basic build tool set to match buildd
                apt-get -y --force-yes install {build_pkgs}
                # Set up expected /dev entries
                if [ ! -r /dev/stdin ];  then
                    ln -s /proc/self/fd/0 /dev/stdin
                fi
                if [ ! -r /dev/stdout ]; then
                    ln -s /proc/self/fd/1 /dev/stdout
                fi
                if [ ! -r /dev/stderr ]; then
                    ln -s /proc/self/fd/2 /dev/stderr
                fi
                # Clean up
                rm /finish.sh
                apt-get clean
            """).format(build_pkgs=' '.join(build_pkgs)))
        return finish_script

    def _debootstrap(self, components, mount, archive_mirror, ports_mirror):
        if self.native_arch in primary_arches:
            mirror = archive_mirror
        else:
            mirror = ports_mirror
        subprocess.check_call([
            "debootstrap",
            "--arch", self.native_arch,
            "--variant=buildd",
            "--components=%s" % ','.join(components),
            self.series,
            mount,
            mirror,
            ])

    @property
    def framework_base(self):
        if self.framework in framework_base:
            return framework_base[self.framework]
        else:
            return self.framework

    @property
    def full_name(self):
        return "%s-%s-%s" % (self.name, self.framework_base, self.target_arch)

    @property
    def full_session_name(self):
        return "%s-%s" % (self.full_name, self.session)

    @property
    def chroot_config(self):
        return "/etc/schroot/chroot.d/%s" % self.full_name

    def exists(self):
        command = ["schroot", "-c", self.full_name, "-i"]
        with open("/dev/null", "w") as devnull:
            return subprocess.call(
                command, stdout=devnull, stderr=devnull) == 0

    def _make_executable(self, path):
        mode = stat.S_IMODE(os.stat(path).st_mode)
        os.chmod(path, mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

    def _make_cross_package(self, prefix):
        if self.native_arch == self.target_arch:
            return prefix
        else:
            target_tuple = self.dpkg_architecture["DEB_HOST_GNU_TYPE"]
            return "%s-%s" % (prefix, target_tuple)

    def create(self, keep_broken_chroot_on_fail=False):
        if self.exists():
            raise ClickChrootAlreadyExistsException(
                "Chroot %s already exists" % self.full_name)
        components = ["main", "restricted", "universe", "multiverse"]
        mount = "%s/%s" % (self.chroots_dir, self.full_name)
        proxy = None
        if not proxy and "http_proxy" in os.environ:
            proxy = os.environ["http_proxy"]
        if not proxy:
            proxy = subprocess.check_output(
                'unset x; eval "$(apt-config shell x Acquire::HTTP::Proxy)"; \
                 echo "$x"',
                shell=True, universal_newlines=True).strip()
        build_pkgs = [
            # sort alphabetically
            "apt-utils",
            "build-essential",
            "cmake",
            "dpkg-cross",
            "fakeroot",
            "libc-dev:%s" % self.target_arch,
            # build pkg names dynamically
            self._make_cross_package("g++"),
            self._make_cross_package("pkg-config"),
        ]
        for package in extra_packages.get(self.framework_base, []):
            package = package.format(TARGET=self.target_arch)
            build_pkgs.append(package)
        os.makedirs(mount)

        country_code = get_geoip_country_code_prefix()
        archive_mirror = "http://%sarchive.ubuntu.com/ubuntu" % country_code
        ports_mirror = "http://%sports.ubuntu.com/ubuntu-ports" % country_code
        # this doesn't work because we are running this under sudo
        if 'DEBOOTSTRAP_MIRROR' in os.environ:
            archive_mirror = os.environ['DEBOOTSTRAP_MIRROR']
        self._debootstrap(components, mount, archive_mirror, ports_mirror)
        sources = generate_sources(self.series, self.native_arch,
                                   self.target_arch,
                                   archive_mirror, ports_mirror,
                                   ' '.join(components))
        with open("%s/etc/apt/sources.list" % mount, "w") as sources_list:
            for line in sources:
                print(line, file=sources_list)
        shutil.copy2("/etc/localtime", "%s/etc/" % mount)
        shutil.copy2("/etc/timezone", "%s/etc/" % mount)
        self._generate_chroot_config(mount)
        daemon_policy = self._generate_daemon_policy(mount)
        self._make_executable(daemon_policy)
        initctl = "%s/sbin/initctl" % mount
        if os.path.exists(initctl):
            os.remove(initctl)
            os.symlink("%s/bin/true" % mount, initctl)
        self._generate_apt_proxy_file(mount, proxy)
        finish_script = self._generate_finish_script(mount, build_pkgs)
        self._make_executable(finish_script)
        command = ["/finish.sh"]
        ret_code = self.maint(*command)
        if ret_code != 0 and not keep_broken_chroot_on_fail:
            # cleanup on failure
            self.destroy()
            raise ClickChrootException(
                "Failed to create chroot '{}' (exit status {})".format(
                    self.full_name, ret_code))
        return ret_code

    def run(self, *args):
        if not self.exists():
            raise ClickChrootDoesNotExistException(
                "Chroot %s does not exist" % self.full_name)
        command = ["schroot", "-c"]
        if self.session:
            command.extend([self.full_session_name, "--run-session"])
        else:
            command.append(self.full_name)
        command.extend(["--", "env"])
        for key, value in self.dpkg_architecture.items():
            command.append("%s=%s" % (key, value))
        command.extend(args)
        ret = subprocess.call(command)
        if ret == 0:
            return 0
        else:
            print("Command returned %d: %s" % (ret, shell_escape(command)),
                  file=sys.stderr)
            return ret

    def maint(self, *args):
        command = ["schroot", "-u", "root", "-c"]
        if self.session:
            command.extend([self.full_session_name, "--run-session"])
        else:
            command.append("source:%s" % self.full_name)
        command.append("--")
        command.extend(args)
        ret = subprocess.call(command)
        if ret == 0:
            return 0
        else:
            print("Command returned %d: %s" % (ret, shell_escape(command)),
                  file=sys.stderr)
            return ret

    def install(self, *pkgs):
        if not self.exists():
            raise ClickChrootDoesNotExistException(
                "Chroot %s does not exist" % self.full_name)
        ret = self.update()
        if ret != 0:
            return ret
        command = ["apt-get", "install", "--yes"]
        command.extend(pkgs)
        ret = self.maint(*command)
        if ret != 0:
            return ret
        return self.clean()

    def clean(self):
        command = ["apt-get", "clean"]
        return self.maint(*command)

    def update(self):
        command = ["apt-get", "update", "--yes"]
        return self.maint(*command)

    def upgrade(self):
        if not self.exists():
            raise ClickChrootDoesNotExistException(
                "Chroot %s does not exist" % self.full_name)
        ret = self.update()
        if ret != 0:
            return ret
        command = ["apt-get", "dist-upgrade", "--yes"]
        ret = self.maint(*command)
        if ret != 0:
            return ret
        return self.clean()

    def destroy(self):
        # remove config
        if os.path.exists(self.chroot_config):
            os.remove(self.chroot_config)
        # find all schroot mount points, this is actually quite complicated
        mount_dir = os.path.abspath(
            os.path.join(self.chroots_dir, "..", "mount"))
        needle = os.path.join(mount_dir, self.full_name)
        all_mounts = []
        with open("/proc/mounts") as f:
            for line in f.readlines():
                mp = line.split()[1]
                if mp.startswith(needle):
                    all_mounts.append(mp)
        # reverse order is important in case of submounts
        for mp in sorted(all_mounts, key=len, reverse=True):
            subprocess.call(["umount", mp])
        # now remove the rest
        chroot_dir = "%s/%s" % (self.chroots_dir, self.full_name)
        if os.path.exists(chroot_dir):
            shutil.rmtree(chroot_dir)
        return 0

    def begin_session(self):
        if not self.exists():
            raise ClickChrootDoesNotExistException(
                "Chroot %s does not exist" % self.full_name)
        command = ["schroot", "-c", self.full_name, "--begin-session",
                   "--session-name", self.full_session_name]
        subprocess.check_call(command)
        return 0

    def end_session(self):
        if not self.exists():
            raise ClickChrootDoesNotExistException(
                "Chroot %s does not exist" % self.full_name)
        command = ["schroot", "-c", self.full_session_name, "--end-session"]
        subprocess.check_call(command)
        return 0



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""List installed Click packages."""

from __future__ import print_function

import json
from optparse import OptionParser
import sys

from gi.repository import Click

from click_package.json_helpers import json_array_to_python


def list_packages(options):
    db = Click.DB()
    db.read(db_dir=None)
    if options.root is not None:
        db.add(options.root)
    if options.all:
        return json_array_to_python(db.get_manifests(all_versions=True))
    else:
        registry = Click.User.for_user(db, name=options.user)
        return json_array_to_python(registry.get_manifests())


def run(argv):
    parser = OptionParser("%prog list [options]")
    parser.add_option(
        "--root", metavar="PATH", help="look for additional packages in PATH")
    parser.add_option(
        "--all", default=False, action="store_true",
        help="list all installed packages")
    parser.add_option(
        "--user", metavar="USER",
        help="list packages registered by USER (if you have permission)")
    parser.add_option(
        "--manifest", default=False, action="store_true",
        help="format output as a JSON array of manifests")
    options, _ = parser.parse_args(argv)
    json_output = list_packages(options)
    if options.manifest:
        json.dump(
            json_output, sys.stdout, ensure_ascii=False, sort_keys=True,
            indent=4, separators=(",", ": "))
        print()
    else:
        for manifest in json_output:
            print("%s\t%s" % (manifest["name"], manifest["version"]))
    return 0



# Copyright (C) 2014 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for click_package.query."""

from __future__ import print_function
__all__ = [
    'TestQuery',
    ]


import os

from gi.repository import Click

from click_package.tests.helpers import TestCase, touch


class TestQuery(TestCase):
    def setUp(self):
        super(TestQuery, self).setUp()
        self.use_temp_dir()

    def test_find_package_directory_missing(self):
        path = os.path.join(self.temp_dir, "nonexistent")
        self.assertRaisesQueryError(
            Click.QueryError.PATH, Click.find_package_directory, path)

    def test_find_package_directory(self):
        info = os.path.join(self.temp_dir, ".click", "info")
        path = os.path.join(self.temp_dir, "file")
        Click.ensuredir(info)
        touch(path)
        pkgdir = Click.find_package_directory(path)
        self.assertEqual(self.temp_dir, pkgdir)

    def test_find_package_directory_outside(self):
        self.assertRaisesQueryError(
            Click.QueryError.NO_PACKAGE_DIR, Click.find_package_directory,
            "/bin")



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for the click signature checking."""

import copy
import os
import shutil
import subprocess
import tarfile
from textwrap import dedent

import apt

from click_package import osextras
from .helpers import (
    require_root,
    ClickTestCase,
)


def makedirs(path):
    try:
        os.makedirs(path)
    except OSError:
        pass


def get_keyid_from_gpghome(gpg_home):
    """Return the public keyid of a given gpg home dir"""
    output = subprocess.check_output(
        ["gpg", "--home", gpg_home, "--list-keys", "--with-colons"],
        universal_newlines=True)
    for line in output.splitlines():
        if not line.startswith("pub:"):
            continue
        return line.split(":")[4]
    raise ValueError("Cannot find public key in output: '%s'" % output)


class Debsigs:
    """Tiny wrapper around the debsigs CLI"""
    def __init__(self, gpghome, keyid):
        self.keyid = keyid
        self.gpghome = gpghome
        self.policy = "/etc/debsig/policies/%s/generic.pol" % self.keyid

    def sign(self, filepath, signature_type="origin"):
        """Sign the click at filepath"""
        env = copy.copy(os.environ)
        env["GNUPGHOME"] = os.path.abspath(self.gpghome)
        try:
            subprocess.check_call(
                ["debsigs",
                 "--sign=%s" % signature_type,
                 "--default-key=%s" % self.keyid,
                 filepath], env=env)
        finally:
            if osextras.find_on_path("gpgconf"):
                subprocess.call(["gpgconf", "--kill", "gpg-agent"])

    def install_signature_policy(self):
        """Install/update the system-wide signature policy"""
        if apt.Cache()["debsig-verify"].installed >= "0.15":
            debsig_xmlns = "https://www.debian.org/debsig/1.0/"
        else:
            debsig_xmlns = "http://www.debian.org/debsig/1.0/"
        xmls = dedent("""\
        <?xml version="1.0"?>
        <!DOCTYPE Policy SYSTEM "{debsig_xmlns}policy.dtd">
        <Policy xmlns="{debsig_xmlns}">

        <Origin Name="test-origin" id="{keyid}" Description="Example policy"/>
        <Selection>
        <Required Type="origin" File="{filename}" id="{keyid}"/>
        </Selection>

        <Verification>
        <Required Type="origin" File="{filename}" id="{keyid}"/>
        </Verification>
        </Policy>
        """.format(
            debsig_xmlns=debsig_xmlns, keyid=self.keyid,
            filename="origin.pub"))
        makedirs(os.path.dirname(self.policy))
        with open(self.policy, "w") as f:
            f.write(xmls)
        self.pubkey_path = (
            "/usr/share/debsig/keyrings/%s/origin.pub" % self.keyid)
        makedirs(os.path.dirname(self.pubkey_path))
        shutil.copy(
            os.path.join(self.gpghome, "pubring.gpg"), self.pubkey_path)

    def uninstall_signature_policy(self):
        # FIXME: update debsig-verify so that it can work from a different
        #        root than "/" so that the tests do not have to use the
        #        system root
        os.remove(self.policy)
        os.remove(self.pubkey_path)


class ClickSignaturesTestCase(ClickTestCase):

    @classmethod
    def setUpClass(cls):
        super(ClickSignaturesTestCase, cls).setUpClass()
        require_root()

    def assertClickNoSignatureError(self, cmd_args):
        with self.assertRaises(subprocess.CalledProcessError) as cm:
            output = subprocess.check_output(
                [self.click_binary] + cmd_args,
                stderr=subprocess.STDOUT, universal_newlines=True)
        output = cm.exception.output
        expected_error_message = ("debsig: Origin Signature check failed. "
                                  "This deb might not be signed.")
        self.assertIn(expected_error_message, output)

    def assertClickInvalidSignatureError(self, cmd_args):
        with self.assertRaises(subprocess.CalledProcessError) as cm:
            output = subprocess.check_output(
                [self.click_binary] + cmd_args,
                stderr=subprocess.STDOUT, universal_newlines=True)
            print(output)

        output = cm.exception.output
        expected_error_message = "Signature verification error: "
        self.assertIn(expected_error_message, output)


class TestSignatureVerificationNoSignature(ClickSignaturesTestCase):

    @classmethod
    def setUpClass(cls):
        super(TestSignatureVerificationNoSignature, cls).setUpClass()
        require_root()

    def test_debsig_verify_no_sig(self):
        name = "org.example.debsig-no-sig"
        path_to_click = self._make_click(name, framework="")
        self.assertClickNoSignatureError(["verify", path_to_click])

    def test_debsig_install_no_sig(self):
        name = "org.example.debsig-no-sig"
        path_to_click = self._make_click(name, framework="")
        self.assertClickNoSignatureError(["install", path_to_click])

    def test_debsig_install_can_install_with_sig_override(self):
        name = "org.example.debsig-no-sig"
        path_to_click = self._make_click(name, framework="")
        user = os.environ.get("USER", "root")
        subprocess.check_call(
            [self.click_binary, "install",
             "--allow-unauthenticated", "--user=%s" % user,
             path_to_click])
        self.addCleanup(
            subprocess.call, [self.click_binary, "unregister",
                              "--user=%s" % user, name])


class TestSignatureVerification(ClickSignaturesTestCase):

    @classmethod
    def setUpClass(cls):
        super(TestSignatureVerification, cls).setUpClass()
        require_root()

    def setUp(self):
        super(TestSignatureVerification, self).setUp()
        self.user = os.environ.get("USER", "root")
        # the valid origin keyring
        self.datadir = os.path.join(os.path.dirname(__file__), "data")
        origin_keyring_dir = os.path.abspath(
            os.path.join(self.datadir, "origin-keyring"))
        gpghome = self.make_gpghome(origin_keyring_dir)
        keyid = get_keyid_from_gpghome(gpghome)
        self.debsigs = Debsigs(gpghome, keyid)
        self.debsigs.install_signature_policy()

    def tearDown(self):
        self.debsigs.uninstall_signature_policy()

    def make_gpghome(self, source):
        gpghome = os.path.join(self.temp_dir, "gnupg")
        if os.path.exists(gpghome):
            shutil.rmtree(gpghome)
        shutil.copytree(source, gpghome)
        os.chmod(gpghome, 0o700)
        return gpghome

    def test_debsig_install_valid_signature(self):
        name = "org.example.debsig-valid-sig"
        path_to_click = self._make_click(name, framework="")
        self.debsigs.sign(path_to_click)
        subprocess.call(["cp", path_to_click, os.path.join("/home/cjwatson/src/ubuntu/click/click", os.path.basename(path_to_click))])
        subprocess.check_call(
            [self.click_binary, "install",
             "--user=%s" % self.user,
             path_to_click])
        self.addCleanup(
            subprocess.call, [self.click_binary, "unregister",
                              "--user=%s" % self.user, name])
        output = subprocess.check_output(
            [self.click_binary, "list", "--user=%s" % self.user],
            universal_newlines=True)
        self.assertIn(name, output)

    def test_debsig_install_signature_not_in_keyring(self):
        name = "org.example.debsig-no-keyring-sig"
        path_to_click = self._make_click(name, framework="")
        evil_keyring_dir = os.path.join(self.datadir, "evil-keyring")
        gpghome = self.make_gpghome(evil_keyring_dir)
        keyid = get_keyid_from_gpghome(gpghome)
        debsig_bad = Debsigs(gpghome, keyid)
        debsig_bad.sign(path_to_click)
        # and ensure its really not there
        self.assertClickInvalidSignatureError(["install", path_to_click])
        output = subprocess.check_output(
            [self.click_binary, "list", "--user=%s" % self.user],
            universal_newlines=True)
        self.assertNotIn(name, output)

    def test_debsig_install_not_a_signature(self):
        name = "org.example.debsig-invalid-sig"
        path_to_click = self._make_click(name, framework="")
        invalid_sig = os.path.join(self.temp_dir, "_gpgorigin")
        with open(invalid_sig, "w") as f:
            f.write("no-valid-signature")
        # add a invalid sig
        subprocess.check_call(["ar", "-r", path_to_click, invalid_sig])
        self.assertClickInvalidSignatureError(["install", path_to_click])
        output = subprocess.check_output(
            [self.click_binary, "list", "--user=%s" % self.user],
            universal_newlines=True)
        self.assertNotIn(name, output)

    def test_debsig_install_signature_altered_click(self):
        def modify_ar_member(member):
            subprocess.check_call(
                ["ar", "-x", path_to_click, "control.tar.gz"],
                cwd=self.temp_dir)
            altered_member = os.path.join(self.temp_dir, member)
            with open(altered_member, "ba") as f:
                f.write(b"\0")
            subprocess.check_call(["ar", "-r", path_to_click, altered_member])

        # ensure that all members we care about are checked by debsig-verify
        for member in ["control.tar.gz", "data.tar.gz", "debian-binary"]:
            name = "org.example.debsig-altered-click"
            path_to_click = self._make_click(name, framework="")
            self.debsigs.sign(path_to_click)
            modify_ar_member(member)
            self.assertClickInvalidSignatureError(["install", path_to_click])
            output = subprocess.check_output(
                [self.click_binary, "list", "--user=%s" % self.user],
                universal_newlines=True)
            self.assertNotIn(name, output)

    def make_nasty_data_tar(self, compression):
        new_data_tar = os.path.join(self.temp_dir, "data.tar." + compression)
        evilfile = os.path.join(self.temp_dir, "README.evil")
        with open(evilfile, "w") as f:
            f.write("I am a nasty README")
        with tarfile.open(new_data_tar, "w:"+compression) as tar:
            tar.add(evilfile)
        return new_data_tar

    def test_debsig_install_signature_injected_data_tar(self):
        name = "org.example.debsig-injected-data-click"
        path_to_click = self._make_click(name, framework="")
        self.debsigs.sign(path_to_click)
        new_data = self.make_nasty_data_tar("bz2")
        # insert before the real data.tar.gz and ensure this is caught
        # NOTE: that right now this will not be caught by debsig-verify
        #        but later in audit() by debian.debfile.DebFile()
        subprocess.check_call(["ar",
                               "-r",
                               "-b", "data.tar.gz",
                               path_to_click,
                               new_data])
        output = subprocess.check_output(
            ["ar", "-t", path_to_click], universal_newlines=True)
        self.assertEqual(output.splitlines(),
                         ["debian-binary",
                          "_click-binary",
                          "control.tar.gz",
                          "data.tar.bz2",
                          "data.tar.gz",
                          "_gpgorigin"])
        with self.assertRaises(subprocess.CalledProcessError):
            output = subprocess.check_output(
                [self.click_binary, "install", path_to_click],
                stderr=subprocess.STDOUT, universal_newlines=True)
        output = subprocess.check_output(
            [self.click_binary, "list", "--user=%s" % self.user],
            universal_newlines=True)
        self.assertNotIn(name, output)

    def test_debsig_install_signature_replaced_data_tar(self):
        name = "org.example.debsig-replaced-data-click"
        path_to_click = self._make_click(name, framework="")
        self.debsigs.sign(path_to_click)
        new_data = self.make_nasty_data_tar("bz2")
        # replace data.tar.gz with data.tar.bz2 and ensure this is caught
        subprocess.check_call(["ar",
                               "-d",
                               path_to_click,
                               "data.tar.gz",
                               ])
        subprocess.check_call(["ar",
                               "-r",
                               path_to_click,
                               new_data])
        output = subprocess.check_output(
            ["ar", "-t", path_to_click], universal_newlines=True)
        self.assertEqual(output.splitlines(),
                         ["debian-binary",
                          "_click-binary",
                          "control.tar.gz",
                          "_gpgorigin",
                          "data.tar.bz2",
                          ])
        with self.assertRaises(subprocess.CalledProcessError) as cm:
            output = subprocess.check_output(
                [self.click_binary, "install", path_to_click],
                stderr=subprocess.STDOUT, universal_newlines=True)
        self.assertIn("Signature verification error", cm.exception.output)
        output = subprocess.check_output(
            [self.click_binary, "list", "--user=%s" % self.user],
            universal_newlines=True)
        self.assertNotIn(name, output)

    def test_debsig_install_signature_prepend_sig(self):
        # this test is probably not really needed, it tries to trick
        # the system by prepending a valid signature that is not
        # in the keyring. But given that debsig-verify only reads
        # the first packet of any given _gpg$foo signature it's
        # equivalent to test_debsig_install_signature_not_in_keyring test
        name = "org.example.debsig-replaced-data-prepend-sig-click"
        path_to_click = self._make_click(name, framework="")
        self.debsigs.sign(path_to_click)
        new_data = self.make_nasty_data_tar("gz")
        # replace data.tar.gz
        subprocess.check_call(["ar",
                               "-r",
                               path_to_click,
                               new_data,
                               ])
        # get previous good _gpgorigin for the old data
        subprocess.check_call(
            ["ar", "-x", path_to_click, "_gpgorigin"], cwd=self.temp_dir)
        with open(os.path.join(self.temp_dir, "_gpgorigin"), "br") as f:
            good_gpg_origin = f.read()
        # and append a valid signature from a non-keyring key
        evil_keyring_dir = os.path.join(self.datadir, "evil-keyring")
        gpghome = self.make_gpghome(evil_keyring_dir)
        debsig_bad = Debsigs(gpghome, "18B38B9AC1B67A0D")
        debsig_bad.sign(path_to_click)
        subprocess.check_call(
            ["ar", "-x", path_to_click, "_gpgorigin"], cwd=self.temp_dir)
        with open(os.path.join(self.temp_dir, "_gpgorigin"), "br") as f:
            evil_gpg_origin = f.read()
        with open(os.path.join(self.temp_dir, "_gpgorigin"), "wb") as f:
            f.write(evil_gpg_origin)
            f.write(good_gpg_origin)
        subprocess.check_call(
            ["ar", "-r", path_to_click, "_gpgorigin"], cwd=self.temp_dir)
        # now ensure that the verification fails as well
        with self.assertRaises(subprocess.CalledProcessError) as cm:
            output = subprocess.check_output(
                [self.click_binary, "install", path_to_click],
                stderr=subprocess.STDOUT, universal_newlines=True)
        self.assertIn("Signature verification error", cm.exception.output)
        output = subprocess.check_output(
            [self.click_binary, "list", "--user=%s" % self.user],
            universal_newlines=True)
        self.assertNotIn(name, output)



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for click_package.user."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'TestClickUser',
    ]


import json
import os
import shutil
from textwrap import dedent

from gi.repository import Click, GLib

from click_package.json_helpers import json_array_to_python, json_object_to_python
from click_package.tests.gimock_types import Passwd
from click_package.tests.helpers import (
    TestCase,
    make_installed_click,
    mkfile,
    make_file_with_content,
    touch,
)


class TestClickUser(TestCase):
    def setUp(self):
        super(TestClickUser, self).setUp()
        self.use_temp_dir()
        self.db = Click.DB()
        self.db.add(self.temp_dir)

    def _setUpMultiDB(self):
        self.multi_db = Click.DB()
        self.multi_db.add(os.path.join(self.temp_dir, "custom"))
        self.multi_db.add(os.path.join(self.temp_dir, "click"))
        user_dbs = [
            os.path.join(
                self.multi_db.get(i).props.root, ".click", "users", "user")
            for i in range(self.multi_db.props.size)
        ]
        a_1_0 = os.path.join(self.temp_dir, "custom", "a", "1.0")
        os.makedirs(a_1_0)
        with mkfile(os.path.join(a_1_0, ".click", "info", "a.manifest")) as m:
            json.dump({"name": "a", "version": "1.0",
                       "hooks": {"a-app": {}}}, m)
        b_2_0 = os.path.join(self.temp_dir, "custom", "b", "2.0")
        os.makedirs(b_2_0)
        with mkfile(os.path.join(b_2_0, ".click", "info", "b.manifest")) as m:
            json.dump({"name": "b", "version": "2.0"}, m)
        a_1_1 = os.path.join(self.temp_dir, "click", "a", "1.1")
        os.makedirs(a_1_1)
        with mkfile(os.path.join(a_1_1, ".click", "info", "a.manifest")) as m:
            json.dump({"name": "a", "version": "1.1"}, m)
        c_0_1 = os.path.join(self.temp_dir, "click", "c", "0.1")
        os.makedirs(c_0_1)
        with mkfile(os.path.join(c_0_1, ".click", "info", "c.manifest")) as m:
            json.dump({"name": "c", "version": "0.1"}, m)
        os.makedirs(user_dbs[0])
        os.symlink(a_1_0, os.path.join(user_dbs[0], "a"))
        os.symlink(b_2_0, os.path.join(user_dbs[0], "b"))
        os.makedirs(user_dbs[1])
        os.symlink(a_1_1, os.path.join(user_dbs[1], "a"))
        os.symlink(c_0_1, os.path.join(user_dbs[1], "c"))
        return user_dbs, Click.User.for_user(self.multi_db, "user")

    def test_new_no_db(self):
        with self.run_in_subprocess(
                "click_get_db_dir", "g_get_user_name") as (enter, preloads):
            enter()
            preloads["click_get_db_dir"].side_effect = (
                lambda: self.make_string(self.temp_dir))
            preloads["g_get_user_name"].side_effect = (
                lambda: self.make_string("test-user"))
            db_root = os.path.join(self.temp_dir, "db")
            os.makedirs(db_root)
            with open(os.path.join(self.temp_dir, "db.conf"), "w") as f:
                print("[Click Database]", file=f)
                print("root = %s" % db_root, file=f)
            registry = Click.User.for_user(None, None)
            self.assertEqual(
                os.path.join(db_root, ".click", "users", "test-user"),
                registry.get_overlay_db())

    def test_new_db_not_directory(self):
        with self.run_in_subprocess(
                "click_get_db_dir", "g_get_user_name") as (enter, preloads):
            enter()
            path = os.path.join(self.temp_dir, "file")
            touch(path)
            preloads["click_get_db_dir"].side_effect = (
                lambda: self.make_string(path))
            self.assertRaisesFileError(
                GLib.FileError.NOTDIR, Click.User.for_user, None, None)

    def test_get_overlay_db(self):
        self.assertEqual(
            os.path.join(self.temp_dir, ".click", "users", "user"),
            Click.User.for_user(self.db, "user").get_overlay_db())

    def test_ensure_db_ownership(self):
        # getpwnam results are cached properly, in a way that doesn't fail
        # due to confusion with getpwnam returning a pointer to a static
        # buffer.
        with self.run_in_subprocess(
                "chown", "geteuid", "getpwnam") as (enter, preloads):
            enter()
            preloads["geteuid"].return_value = 0
            getpwnam_result = Passwd()

            def getpwnam_side_effect(name):
                if name == b"clickpkg":
                    getpwnam_result.pw_uid = 1
                    getpwnam_result.pw_gid = 1
                else:
                    getpwnam_result.pw_uid = 2
                    getpwnam_result.pw_gid = 2
                return self.make_pointer(getpwnam_result)

            preloads["getpwnam"].side_effect = getpwnam_side_effect
            registry = Click.User.for_user(self.db, "user")
            os.makedirs(os.path.join(self.temp_dir, "a", "1.0"))
            click_dir = os.path.join(self.temp_dir, ".click")

            registry.set_version("a", "1.0")
            self.assertEqual(3, preloads["chown"].call_count)
            preloads["chown"].assert_any_call(click_dir.encode(), 1, 1)
            preloads["chown"].assert_any_call(
                os.path.join(click_dir, "users").encode(), 1, 1)
            preloads["chown"].assert_any_call(
                os.path.join(click_dir, "users", "user").encode(), 2, 2)

            # Try again, now that both password file entries should be
            # cached.
            shutil.rmtree(os.path.join(self.temp_dir, ".click"))
            preloads["chown"].reset_mock()
            registry.set_version("a", "1.0")
            self.assertEqual(3, preloads["chown"].call_count)
            preloads["chown"].assert_any_call(click_dir.encode(), 1, 1)
            preloads["chown"].assert_any_call(
                os.path.join(click_dir, "users").encode(), 1, 1)
            preloads["chown"].assert_any_call(
                os.path.join(click_dir, "users", "user").encode(), 2, 2)

    def test_ensure_db_mkdir_fails(self):
        with self.run_in_subprocess("mkdir") as (enter, preloads):
            enter()
            preloads["mkdir"].return_value = -1
            registry = Click.User.for_user(self.db, "user")
            self.assertRaisesUserError(
                Click.UserError.CREATE_DB, registry.set_version, "a", "1.0")

    def test_ensure_db_chown_fails(self):
        with self.run_in_subprocess(
                "chown", "geteuid", "getpwnam") as (enter, preloads):
            enter()
            preloads["geteuid"].return_value = 0
            getpwnam_result = Passwd()

            def getpwnam_side_effect(name):
                if name == b"clickpkg":
                    getpwnam_result.pw_uid = 1
                    getpwnam_result.pw_gid = 1
                else:
                    getpwnam_result.pw_uid = 2
                    getpwnam_result.pw_gid = 2
                return self.make_pointer(getpwnam_result)

            preloads["getpwnam"].side_effect = getpwnam_side_effect
            preloads["chown"].return_value = -1
            registry = Click.User.for_user(self.db, "user")
            self.assertRaisesUserError(
                Click.UserError.CHOWN_DB, registry.set_version, "a", "1.0")

    def test_ensure_db_getpwnam_fails(self):
        with self.run_in_subprocess(
                "geteuid", "getpwnam") as (enter, preloads):
            enter()
            preloads["geteuid"].return_value = 0
            preloads["getpwnam"].return_value = None
            registry = Click.User.for_user(self.db, "user")
            self.assertRaisesUserError(
                Click.UserError.GETPWNAM, registry.set_version, "a", "1.0")

    def test_get_package_names_missing(self):
        db = Click.DB()
        db.add(os.path.join(self.temp_dir, "nonexistent"))
        registry = Click.User.for_user(db, None)
        self.assertEqual([], list(registry.get_package_names()))

    def test_get_package_names(self):
        registry = Click.User.for_user(self.db, "user")
        os.makedirs(registry.get_overlay_db())
        os.symlink("/1.0", os.path.join(registry.get_overlay_db(), "a"))
        os.symlink("/1.1", os.path.join(registry.get_overlay_db(), "b"))
        self.assertCountEqual(["a", "b"], list(registry.get_package_names()))

    def test_get_package_names_multiple_root(self):
        _, registry = self._setUpMultiDB()
        self.assertCountEqual(
            ["a", "b", "c"], list(registry.get_package_names()))

    def test_get_version_missing(self):
        registry = Click.User.for_user(self.db, "user")
        self.assertRaisesUserError(
            Click.UserError.NO_SUCH_PACKAGE, registry.get_version, "a")
        self.assertFalse(registry.has_package_name("a"))

    def test_get_version(self):
        registry = Click.User.for_user(self.db, "user")
        os.makedirs(registry.get_overlay_db())
        os.symlink("/1.0", os.path.join(registry.get_overlay_db(), "a"))
        self.assertEqual("1.0", registry.get_version("a"))
        self.assertTrue(registry.has_package_name("a"))

    def test_get_version_multiple_root(self):
        _, registry = self._setUpMultiDB()
        self.assertEqual("1.1", registry.get_version("a"))
        self.assertEqual("2.0", registry.get_version("b"))
        self.assertEqual("0.1", registry.get_version("c"))
        self.assertTrue(registry.has_package_name("a"))
        self.assertTrue(registry.has_package_name("b"))
        self.assertTrue(registry.has_package_name("c"))

    def test_set_version_missing_target(self):
        registry = Click.User.for_user(self.db, "user")
        self.assertRaisesDatabaseError(
            Click.DatabaseError.DOES_NOT_EXIST,
            registry.set_version, "a", "1.0")

    def test_set_version_missing(self):
        registry = Click.User.for_user(self.db, "user")
        os.makedirs(os.path.join(self.temp_dir, "a", "1.0"))
        registry.set_version("a", "1.0")
        path = os.path.join(registry.get_overlay_db(), "a")
        self.assertTrue(os.path.islink(path))
        self.assertEqual(
            os.path.join(self.temp_dir, "a", "1.0"), os.readlink(path))

    def test_set_version_changed(self):
        registry = Click.User.for_user(self.db, "user")
        os.makedirs(registry.get_overlay_db())
        path = os.path.join(registry.get_overlay_db(), "a")
        os.symlink("/1.0", path)
        os.makedirs(os.path.join(self.temp_dir, "a", "1.1"))
        registry.set_version("a", "1.1")
        self.assertTrue(os.path.islink(path))
        self.assertEqual(
            os.path.join(self.temp_dir, "a", "1.1"), os.readlink(path))

    def test_set_version_multiple_root(self):
        user_dbs, registry = self._setUpMultiDB()

        os.makedirs(os.path.join(self.multi_db.get(1).props.root, "a", "1.2"))
        registry.set_version("a", "1.2")
        a_underlay = os.path.join(user_dbs[0], "a")
        a_overlay = os.path.join(user_dbs[1], "a")
        self.assertTrue(os.path.islink(a_underlay))
        self.assertEqual(
            os.path.join(self.multi_db.get(0).props.root, "a", "1.0"),
            os.readlink(a_underlay))
        self.assertTrue(os.path.islink(a_overlay))
        self.assertEqual(
            os.path.join(self.multi_db.get(1).props.root, "a", "1.2"),
            os.readlink(a_overlay))

        os.makedirs(os.path.join(self.multi_db.get(1).props.root, "b", "2.1"))
        registry.set_version("b", "2.1")
        b_underlay = os.path.join(user_dbs[0], "b")
        b_overlay = os.path.join(user_dbs[1], "b")
        self.assertTrue(os.path.islink(b_underlay))
        self.assertEqual(
            os.path.join(self.multi_db.get(0).props.root, "b", "2.0"),
            os.readlink(b_underlay))
        self.assertTrue(os.path.islink(b_overlay))
        self.assertEqual(
            os.path.join(self.multi_db.get(1).props.root, "b", "2.1"),
            os.readlink(b_overlay))

        os.makedirs(os.path.join(self.multi_db.get(1).props.root, "c", "0.2"))
        registry.set_version("c", "0.2")
        c_underlay = os.path.join(user_dbs[0], "c")
        c_overlay = os.path.join(user_dbs[1], "c")
        self.assertFalse(os.path.islink(c_underlay))
        self.assertTrue(os.path.islink(c_overlay))
        self.assertEqual(
            os.path.join(self.multi_db.get(1).props.root, "c", "0.2"),
            os.readlink(c_overlay))

        os.makedirs(os.path.join(self.multi_db.get(1).props.root, "d", "3.0"))
        registry.set_version("d", "3.0")
        d_underlay = os.path.join(user_dbs[0], "d")
        d_overlay = os.path.join(user_dbs[1], "d")
        self.assertFalse(os.path.islink(d_underlay))
        self.assertTrue(os.path.islink(d_overlay))
        self.assertEqual(
            os.path.join(self.multi_db.get(1).props.root, "d", "3.0"),
            os.readlink(d_overlay))

    def test_set_version_restore_to_underlay(self):
        user_dbs, registry = self._setUpMultiDB()
        a_underlay = os.path.join(user_dbs[0], "a")
        a_overlay = os.path.join(user_dbs[1], "a")

        # Initial state: 1.0 in underlay, 1.1 in overlay.
        self.assertTrue(os.path.islink(a_underlay))
        self.assertEqual(
            os.path.join(self.multi_db.get(0).props.root, "a", "1.0"),
            os.readlink(a_underlay))
        self.assertTrue(os.path.islink(a_overlay))
        self.assertEqual(
            os.path.join(self.multi_db.get(1).props.root, "a", "1.1"),
            os.readlink(a_overlay))

        # Setting to 1.0 (version in underlay) removes overlay link.
        registry.set_version("a", "1.0")
        self.assertTrue(os.path.islink(a_underlay))
        self.assertEqual(
            os.path.join(self.multi_db.get(0).props.root, "a", "1.0"),
            os.readlink(a_underlay))
        self.assertFalse(os.path.islink(a_overlay))

    def test_remove_missing(self):
        registry = Click.User.for_user(self.db, "user")
        self.assertRaisesUserError(
            Click.UserError.NO_SUCH_PACKAGE, registry.remove, "a")

    def test_remove(self):
        registry = Click.User.for_user(self.db, "user")
        os.makedirs(registry.get_overlay_db())
        path = os.path.join(registry.get_overlay_db(), "a")
        os.symlink("/1.0", path)
        registry.remove("a")
        self.assertFalse(os.path.exists(path))

    def test_remove_multiple_root(self):
        user_dbs, registry = self._setUpMultiDB()
        registry.remove("a")
        self.assertFalse(os.path.exists(os.path.join(user_dbs[1], "a")))
        # Exposed underlay.
        self.assertEqual("1.0", registry.get_version("a"))
        registry.remove("b")
        # Hidden.
        self.assertEqual(
            "@hidden", os.readlink(os.path.join(user_dbs[1], "b")))
        self.assertFalse(registry.has_package_name("b"))
        registry.remove("c")
        self.assertFalse(os.path.exists(os.path.join(user_dbs[1], "c")))
        self.assertFalse(registry.has_package_name("c"))
        self.assertRaisesUserError(
            Click.UserError.NO_SUCH_PACKAGE, registry.remove, "d")

    def test_remove_multiple_root_creates_overlay_directory(self):
        multi_db = Click.DB()
        multi_db.add(os.path.join(self.temp_dir, "preinstalled"))
        multi_db.add(os.path.join(self.temp_dir, "click"))
        user_dbs = [
            os.path.join(multi_db.get(i).props.root, ".click", "users", "user")
            for i in range(multi_db.props.size)
        ]
        a_1_0 = os.path.join(self.temp_dir, "preinstalled", "a", "1.0")
        os.makedirs(a_1_0)
        os.makedirs(user_dbs[0])
        os.symlink(a_1_0, os.path.join(user_dbs[0], "a"))
        self.assertFalse(os.path.exists(user_dbs[1]))
        registry = Click.User.for_user(multi_db, "user")
        self.assertEqual("1.0", registry.get_version("a"))
        registry.remove("a")
        self.assertFalse(registry.has_package_name("a"))
        self.assertEqual(
            "@hidden", os.readlink(os.path.join(user_dbs[1], "a")))

    def test_get_path(self):
        registry = Click.User.for_user(self.db, "user")
        os.makedirs(os.path.join(self.temp_dir, "a", "1.0"))
        registry.set_version("a", "1.0")
        self.assertEqual(
            os.path.join(registry.get_overlay_db(), "a"),
            registry.get_path("a"))

    def test_get_path_multiple_root(self):
        user_dbs, registry = self._setUpMultiDB()
        self.assertEqual(
            os.path.join(user_dbs[1], "a"), registry.get_path("a"))
        self.assertEqual(
            os.path.join(user_dbs[0], "b"), registry.get_path("b"))
        self.assertEqual(
            os.path.join(user_dbs[1], "c"), registry.get_path("c"))
        self.assertRaisesUserError(
            Click.UserError.NO_SUCH_PACKAGE, registry.get_path, "d")

    def test_get_manifest(self):
        registry = Click.User.for_user(self.db, "user")
        manifest_path = os.path.join(
            self.temp_dir, "a", "1.0", ".click", "info", "a.manifest")
        manifest_obj = {"name": "a", "version": "1.0"}
        with mkfile(manifest_path) as manifest:
            json.dump(manifest_obj, manifest)
        manifest_obj["_directory"] = os.path.join(
            registry.get_overlay_db(), "a")
        manifest_obj["_removable"] = 1
        registry.set_version("a", "1.0")
        self.assertEqual(
            manifest_obj, json_object_to_python(registry.get_manifest("a")))
        self.assertEqual(
            manifest_obj, json.loads(registry.get_manifest_as_string("a")))

    def test_get_manifest_multiple_root(self):
        user_dbs, registry = self._setUpMultiDB()
        expected_a = {
            "name": "a",
            "version": "1.1",
            "_directory": os.path.join(user_dbs[1], "a"),
            "_removable": 1,
        }
        self.assertEqual(
            expected_a, json_object_to_python(registry.get_manifest("a")))
        self.assertEqual(
            expected_a, json.loads(registry.get_manifest_as_string("a")))
        expected_b = {
            "name": "b",
            "version": "2.0",
            "_directory": os.path.join(user_dbs[0], "b"),
            "_removable": 1,
        }
        self.assertEqual(
            expected_b, json_object_to_python(registry.get_manifest("b")))
        self.assertEqual(
            expected_b, json.loads(registry.get_manifest_as_string("b")))
        expected_c = {
            "name": "c",
            "version": "0.1",
            "_directory": os.path.join(user_dbs[1], "c"),
            "_removable": 1,
        }
        self.assertEqual(
            expected_c, json_object_to_python(registry.get_manifest("c")))
        self.assertEqual(
            expected_c, json.loads(registry.get_manifest_as_string("c")))
        self.assertRaisesUserError(
            Click.UserError.NO_SUCH_PACKAGE, registry.get_path, "d")

    def test_get_manifests(self):
        registry = Click.User.for_user(self.db, "user")
        a_manifest_path = os.path.join(
            self.temp_dir, "a", "1.0", ".click", "info", "a.manifest")
        a_manifest_obj = {"name": "a", "version": "1.0"}
        with mkfile(a_manifest_path) as a_manifest:
            json.dump(a_manifest_obj, a_manifest)
        registry.set_version("a", "1.0")
        b_manifest_path = os.path.join(
            self.temp_dir, "b", "2.0", ".click", "info", "b.manifest")
        b_manifest_obj = {"name": "b", "version": "2.0"}
        with mkfile(b_manifest_path) as b_manifest:
            json.dump(b_manifest_obj, b_manifest)
        registry.set_version("b", "2.0")
        a_manifest_obj["_directory"] = os.path.join(
            registry.get_overlay_db(), "a")
        a_manifest_obj["_removable"] = 1
        b_manifest_obj["_directory"] = os.path.join(
            registry.get_overlay_db(), "b")
        b_manifest_obj["_removable"] = 1
        self.assertEqual(
            [a_manifest_obj, b_manifest_obj],
            json_array_to_python(registry.get_manifests()))
        self.assertEqual(
            [a_manifest_obj, b_manifest_obj],
            json.loads(registry.get_manifests_as_string()))

    def test_get_manifests_multiple_root(self):
        user_dbs, registry = self._setUpMultiDB()
        a_manifest_obj = {
            "name": "a",
            "version": "1.1",
            "_directory": os.path.join(user_dbs[1], "a"),
            "_removable": 1,
        }
        b_manifest_obj = {
            "name": "b",
            "version": "2.0",
            "_directory": os.path.join(user_dbs[0], "b"),
            "_removable": 1,
        }
        c_manifest_obj = {
            "name": "c",
            "version": "0.1",
            "_directory": os.path.join(user_dbs[1], "c"),
            "_removable": 1,
        }
        self.assertEqual(
            [a_manifest_obj, c_manifest_obj, b_manifest_obj],
            json_array_to_python(registry.get_manifests()))
        self.assertEqual(
            [a_manifest_obj, c_manifest_obj, b_manifest_obj],
            json.loads(registry.get_manifests_as_string()))
        registry.remove("b")
        self.assertEqual(
            "@hidden", os.readlink(os.path.join(user_dbs[1], "b")))
        self.assertEqual(
            [a_manifest_obj, c_manifest_obj],
            json_array_to_python(registry.get_manifests()))
        self.assertEqual(
            [a_manifest_obj, c_manifest_obj],
            json.loads(registry.get_manifests_as_string()))

    def test_is_removable(self):
        registry = Click.User.for_user(self.db, "user")
        os.makedirs(os.path.join(self.temp_dir, "a", "1.0"))
        registry.set_version("a", "1.0")
        self.assertTrue(registry.is_removable("a"))

    def test_is_removable_multiple_root(self):
        user_dbs, registry = self._setUpMultiDB()
        self.assertTrue(registry.is_removable("a"))
        self.assertTrue(registry.is_removable("b"))
        self.assertTrue(registry.is_removable("c"))
        self.assertFalse(registry.is_removable("d"))

    def test_hidden(self):
        user_dbs, registry = self._setUpMultiDB()
        b_overlay = os.path.join(user_dbs[1], "b")

        registry.remove("b")
        self.assertFalse(registry.has_package_name("b"))
        self.assertTrue(os.path.islink(b_overlay))
        self.assertEqual("@hidden", os.readlink(b_overlay))
        self.assertRaisesUserError(
            Click.UserError.HIDDEN_PACKAGE, registry.get_version, "b")
        self.assertRaisesUserError(
            Click.UserError.HIDDEN_PACKAGE, registry.get_path, "b")
        self.assertFalse(registry.is_removable("b"))

        registry.set_version("b", "2.0")
        self.assertTrue(registry.has_package_name("b"))
        self.assertTrue(os.path.islink(b_overlay))
        self.assertEqual(
            os.path.join(self.multi_db.get(0).props.root, "b", "2.0"),
            os.readlink(b_overlay))
        self.assertEqual("2.0", registry.get_version("b"))
        self.assertEqual(b_overlay, registry.get_path("b"))
        self.assertTrue(registry.is_removable("b"))


class StopAppTestCase(TestCase):

    def setUp(self):
        super(StopAppTestCase, self).setUp()
        self.use_temp_dir()
        self.db = Click.DB()
        self.db.add(self.temp_dir)

        # setup fake app_stop
        fake_app_stop = os.path.join(self.temp_dir, "bin", "ubuntu-app-stop")
        self.fake_app_stop_output = os.path.join(
            self.temp_dir, "fake-app-stop.out")
        fake_app_stop_content = dedent("""\
        #!/bin/sh
        echo "$@" >> %s
        """ % self.fake_app_stop_output)
        make_file_with_content(fake_app_stop, fake_app_stop_content, 0o755)
        # its ok to modify env here, click.helpers.TestCase will take care
        # of it
        os.environ["PATH"] = "%s:%s" % (
            os.path.dirname(fake_app_stop), os.environ["PATH"])

    def test_app_stops_on_remove(self):
        make_installed_click(self.db, self.temp_dir, "meep", "2.0",
                             {"hooks": {"a-app": {}}})
        registry = Click.User.for_user(self.db, "user")
        registry.remove("meep")
        # ensure that stop was called with the right app
        with open(self.fake_app_stop_output) as f:
            self.assertEqual("meep_a-app_2.0", f.read().strip())



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for the click install feature."""

import subprocess

from .helpers import (
    require_root,
    ClickTestCase,
)


def add_user(name):
    subprocess.check_call(["useradd", name])
    return name


def del_user(name):
    subprocess.check_call(["userdel", "-r", name])


class TestClickInstall(ClickTestCase):

    @classmethod
    def setUpClass(cls):
        super(TestClickInstall, cls).setUpClass()
        require_root()
        cls.USER_1 = add_user("click-test-user-1")
        cls.USER_2 = add_user("click-test-user-2")

    @classmethod
    def tearDownClass(cls):
        super(TestClickInstall, cls).tearDownClass()
        del_user(cls.USER_1)
        del_user(cls.USER_2)

    def test_install_for_single_user(self):
        name = "foo-1"
        click_pkg = self._make_click(name=name, framework="")
        # install it
        self.click_install(click_pkg, name, self.USER_1)
        # ensure that user-1 has it
        output = subprocess.check_output([
            "sudo", "-u", self.USER_1,
            self.click_binary, "list"], universal_newlines=True)
        self.assertEqual(output, "%s\t1.0\n" % name)
        # but not user-2
        output = subprocess.check_output([
            "sudo", "-u", self.USER_2,
            self.click_binary, "list"], universal_newlines=True)
        self.assertEqual(output, "")
        # and that we can see it with the --user option
        output = subprocess.check_output(
            [self.click_binary, "list", "--user=%s" % self.USER_1],
            universal_newlines=True)
        self.assertEqual(output, "%s\t1.0\n" % name)

    def test_install_for_single_user_and_register(self):
        name = "foo-1"
        click_pkg = self._make_click(name=name, framework="")
        self.click_install(click_pkg, name, self.USER_1)
        # not available for user2
        output = subprocess.check_output([
            "sudo", "-u", self.USER_2,
            self.click_binary, "list"], universal_newlines=True)
        self.assertEqual(output, "")
        # register it
        subprocess.check_call(
            [self.click_binary, "register", "--user=%s" % self.USER_2,
             name, "1.0", ])
        self.addCleanup(self.click_unregister, name, self.USER_2)
        # and ensure its available for user2
        output = subprocess.check_output([
            "sudo", "-u", self.USER_2,
            self.click_binary, "list"], universal_newlines=True)
        self.assertEqual(output, "%s\t1.0\n" % name)

    def test_install_for_all_users(self):
        name = "foo-2"
        click_pkg = self._make_click(name=name, framework="")
        self.click_install(click_pkg, name, "@all")
        # ensure all users see it
        for user in (self.USER_1, self.USER_2):
            output = subprocess.check_output(
                ["sudo", "-u", user, self.click_binary, "list"],
                universal_newlines=True)
            self.assertEqual(output, "%s\t1.0\n" % name)

    def test_pkgdir_after_install(self):
        name = "foo-3"
        click_pkg = self._make_click(name=name, version="1.2", framework="")
        self.click_install(click_pkg, name, "@all")
        # from the path
        output = subprocess.check_output(
            [self.click_binary, "pkgdir",
             "/opt/click.ubuntu.com/%s/1.2/README" % name],
            universal_newlines=True).strip()
        self.assertEqual(output, "/opt/click.ubuntu.com/%s/1.2" % name)
        # now test from the click package name
        output = subprocess.check_output(
            [self.click_binary, "pkgdir", name],
            universal_newlines=True).strip()
        # note that this is different from above
        self.assertEqual(
            output, "/opt/click.ubuntu.com/.click/users/@all/%s" % name)



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Build a Click source package."""

from __future__ import print_function

from optparse import OptionParser
import os
import sys

from click_package.build import ClickBuildError, ClickSourceBuilder


def run(argv):
    parser = OptionParser("%prog buildsource [options] DIRECTORY")
    parser.add_option(
        "-m", "--manifest", metavar="PATH",
        help="read package manifest from PATH")
    parser.add_option(
        "-I", "--ignore", metavar="file-pattern", action='append',
        default=[],
        help="Ignore the given pattern when building the package")
    options, args = parser.parse_args(argv)
    if len(args) < 1:
        parser.error("need directory")
    directory = args[0]
    if not os.path.isdir(directory):
        parser.error('directory "%s" does not exist' % directory)
    if not options.manifest:
        options.manifest = os.path.join(directory, "manifest.json")
    if os.path.isdir(os.path.join(directory, options.manifest)):
        options.manifest = os.path.join(options.manifest, "manifest.json")
    if not os.path.exists(os.path.join(directory, options.manifest)):
        parser.error(
            'directory "%s" does not contain manifest file "%s"' %
            (directory, options.manifest))
    builder = ClickSourceBuilder()
    builder.add_file(directory, "./")
    for ignore in options.ignore:
        builder.add_ignore_pattern(ignore)
    try:
        path = builder.build(".", manifest_path=options.manifest)
    except ClickBuildError as e:
        print(e, file=sys.stderr)
        return 1
    print("Successfully built source package in '%s'." % path)
    return 0



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Show the file-list contents of a Click package file."""

from __future__ import print_function

from optparse import OptionParser
import subprocess


def run(argv):
    parser = OptionParser("%prog contents [options] PATH")
    _, args = parser.parse_args(argv)
    if len(args) < 1:
        parser.error("need file name")
    path = args[0]
    subprocess.check_call(["dpkg-deb", "-c", path])
    return 0



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Basic support for writing ar archive files.

We do things this way so that Click packages can be created with minimal
dependencies (e.g. on non-Ubuntu systems).  No read support is needed, since
Click packages are always installed on systems that have dpkg.

Some method names and general approach come from the tarfile module in
Python's standard library; details of the format come from dpkg.
"""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'ArFile',
    ]

import os
import shutil
import time


class ArFile:
    def __init__(self, name=None, mode="w", fileobj=None):
        if mode != "w":
            raise ValueError("only mode 'w' is supported")
        self.mode = mode
        self.real_mode = "wb"

        if fileobj:
            if name is None and hasattr(fileobj, "name"):
                name = fileobj.name
            if hasattr(fileobj, "mode"):
                if fileobj.mode != "wb":
                    raise ValueError("fileobj must be opened with mode='wb'")
                self._mode = fileobj.mode
            self.opened_fileobj = False
        else:
            fileobj = open(name, self.real_mode)
            self.opened_fileobj = True
        self.name = name
        self.fileobj = fileobj
        self.closed = False

    def close(self):
        if self.opened_fileobj:
            self.fileobj.close()
        self.closed = True

    def _check(self):
        if self.closed:
            raise IOError("ArFile %s is closed" % self.name)

    def __enter__(self):
        self._check()
        return self

    def __exit__(self, *args):
        self.close()

    def add_magic(self):
        self.fileobj.write(b"!<arch>\n")

    def add_header(self, name, size):
        if len(name) > 15:
            raise ValueError("ar member name '%s' length too long" % name)
        if size > 9999999999:
            raise ValueError("ar member size %d too large" % size)
        header = ("%-16s%-12u0     0     100644  %-10d`\n" % (
            name, int(time.time()), size)).encode()
        assert len(header) == 60  # sizeof(struct ar_hdr)
        self.fileobj.write(header)

    def add_data(self, name, data):
        size = len(data)
        self.add_header(name, size)
        self.fileobj.write(data)
        if size & 1:
            self.fileobj.write(b"\n")  # padding

    def add_file(self, name, path):
        with open(path, "rb") as fobj:
            size = os.fstat(fobj.fileno()).st_size
            self.add_header(name, size)
            shutil.copyfileobj(fobj, self.fileobj)
        if size & 1:
            self.fileobj.write(b"\n")  # padding



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Installing Click packages."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'ClickInstaller',
    'ClickInstallerAuditError',
    'ClickInstallerError',
    'ClickInstallerPermissionDenied',
    ]


from functools import partial
import grp
import inspect
import json
import logging
import os
import pwd
import shutil
import stat
import subprocess
import sys
import tempfile
from textwrap import dedent

from contextlib import closing

from debian.debfile import DebFile as _DebFile
from debian.debian_support import Version
from gi.repository import Click

from click_package.paths import preload_path
from click_package.preinst import static_preinst_matches
from click_package.versions import spec_version

from click_package.framework import (
    validate_framework,
    ClickFrameworkInvalid,
)


try:
    _DebFile.close
    DebFile = _DebFile
except AttributeError:
    # Yay!  The Ubuntu 13.04 version of python-debian 0.1.21
    # debian.debfile.DebFile has a .close() method but the PyPI version of
    # 0.1.21 does not.  It's worse than that because DebFile.close() really
    # delegates to DebPart.close() and *that's* missing in the PyPI version.
    # To get that working, we have to reach inside the object and name mangle
    # the attribute.
    class DebFile(_DebFile):
        def close(self):
            self.control._DebPart__member.close()
            self.data._DebPart__member.close()


class DebsigVerifyError(Exception):
    pass


class DebsigVerify:
    """Tiny wrapper around the debsig-verify commandline"""
    # from debsig-verify-0.9/debsigs.h
    DS_SUCCESS = 0
    DS_FAIL_NOSIGS = 10
    DS_FAIL_UNKNOWN_ORIGIN = 11
    DS_FAIL_NOPOLICIES = 12
    DS_FAIL_BADSIG = 13
    DS_FAIL_INTERNAL = 14

    # should be a property, but python does not support support
    # class properties easily
    @classmethod
    def available(cls):
        return Click.find_on_path("debsig-verify")

    @classmethod
    def verify(cls, path, allow_unauthenticated):
        command = ["debsig-verify"] + [path]
        try:
            subprocess.check_output(command, universal_newlines=True)
        except subprocess.CalledProcessError as e:
            if (allow_unauthenticated and
                e.returncode in (DebsigVerify.DS_FAIL_NOSIGS,
                                 DebsigVerify.DS_FAIL_UNKNOWN_ORIGIN,
                                 DebsigVerify.DS_FAIL_NOPOLICIES)):
                logging.warning(
                    "Signature check failed, but installing anyway "
                    "as requested")
            else:
                raise DebsigVerifyError(
                    "Signature verification error: %s" % e.output)
        return True


class ClickInstallerError(Exception):
    pass


class ClickInstallerPermissionDenied(ClickInstallerError):
    pass


class ClickInstallerAuditError(ClickInstallerError):
    pass


class ClickInstaller:
    def __init__(self, db, force_missing_framework=False,
                 allow_unauthenticated=False):
        self.db = db
        self.force_missing_framework = force_missing_framework
        self.allow_unauthenticated = allow_unauthenticated

    def _preload_path(self):
        if "CLICK_PACKAGE_PRELOAD" in os.environ:
            return os.environ["CLICK_PACKAGE_PRELOAD"]
        my_path = inspect.getsourcefile(ClickInstaller)
        preload = os.path.join(
            os.path.dirname(my_path), os.pardir, "preload", ".libs",
            "libclickpreload.so")
        if os.path.exists(preload):
            return os.path.abspath(preload)
        return preload_path

    def _dpkg_architecture(self):
        return subprocess.check_output(
            ["dpkg", "--print-architecture"],
            universal_newlines=True).rstrip("\n")

    def extract(self, path, target):
        command = ["dpkg-deb", "-R", path, target]
        with open(path, "rb") as fd:
            env = dict(os.environ)
            preloads = [self._preload_path()]
            if "LD_PRELOAD" in env:
                preloads.append(env["LD_PRELOAD"])
            env["LD_PRELOAD"] = " ".join(preloads)
            env["CLICK_BASE_DIR"] = target
            env["CLICK_PACKAGE_PATH"] = path
            env["CLICK_PACKAGE_FD"] = str(fd.fileno())
            env.pop("HOME", None)
            kwargs = {}
            if sys.version >= "3.2":
                kwargs["pass_fds"] = (fd.fileno(),)
            subprocess.check_call(command, env=env, **kwargs)

    def audit(self, path, slow=False, check_arch=False):
        # always do the signature check first
        if DebsigVerify.available():
            try:
                DebsigVerify.verify(path, self.allow_unauthenticated)
            except DebsigVerifyError as e:
                raise ClickInstallerAuditError(str(e))
        else:
            logging.warning(
                "debsig-verify not available; cannot check signatures")

        # fail early if the file cannot be opened
        try:
            with closing(DebFile(filename=path)) as package:
                pass
        except Exception as e:
            raise ClickInstallerError("Failed to read %s: %s" % (
                path, str(e)))

        # then perform the audit
        with closing(DebFile(filename=path)) as package:
            control_fields = package.control.debcontrol()

            try:
                click_version = Version(control_fields["Click-Version"])
            except KeyError:
                raise ClickInstallerAuditError("No Click-Version field")
            if click_version > spec_version:
                raise ClickInstallerAuditError(
                    "Click-Version: %s newer than maximum supported version "
                    "%s" % (click_version, spec_version))

            for field in (
                "Pre-Depends", "Depends", "Recommends", "Suggests", "Enhances",
                "Conflicts", "Breaks",
                "Provides",
            ):
                if field in control_fields:
                    raise ClickInstallerAuditError(
                        "%s field is forbidden in Click packages" % field)

            scripts = package.control.scripts()
            if ("preinst" in scripts and
                    static_preinst_matches(scripts["preinst"])):
                scripts.pop("preinst", None)
            if scripts:
                raise ClickInstallerAuditError(
                    "Maintainer scripts are forbidden in Click packages "
                    "(found: %s)" %
                    " ".join(sorted(scripts)))

            if not package.control.has_file("manifest"):
                raise ClickInstallerAuditError("Package has no manifest")
            with package.control.get_file("manifest", encoding="UTF-8") as f:
                manifest = json.load(f)

            try:
                package_name = manifest["name"]
            except KeyError:
                raise ClickInstallerAuditError('No "name" entry in manifest')
            # TODO: perhaps just do full name validation?
            if "/" in package_name:
                raise ClickInstallerAuditError(
                    'Invalid character "/" in "name" entry: %s' % package_name)
            if "_" in package_name:
                raise ClickInstallerAuditError(
                    'Invalid character "_" in "name" entry: %s' % package_name)

            try:
                package_version = manifest["version"]
            except KeyError:
                raise ClickInstallerAuditError(
                    'No "version" entry in manifest')
            # TODO: perhaps just do full version validation?
            if "/" in package_version:
                raise ClickInstallerAuditError(
                    'Invalid character "/" in "version" entry: %s' %
                    package_version)
            if "_" in package_version:
                raise ClickInstallerAuditError(
                    'Invalid character "_" in "version" entry: %s' %
                    package_version)

            try:
                framework = manifest["framework"]
            except KeyError:
                raise ClickInstallerAuditError(
                    'No "framework" entry in manifest')
            try:
                validate_framework(framework, self.force_missing_framework)
            except ClickFrameworkInvalid as e:
                raise ClickInstallerAuditError(str(e))

            if check_arch:
                architecture = manifest.get("architecture", "all")
                if architecture != "all":
                    dpkg_architecture = self._dpkg_architecture()
                    if isinstance(architecture, list):
                        if dpkg_architecture not in architecture:
                            raise ClickInstallerAuditError(
                                'Package architectures "%s" not compatible '
                                'with system architecture "%s"' %
                                (" ".join(architecture), dpkg_architecture))
                    elif architecture != dpkg_architecture:
                        raise ClickInstallerAuditError(
                            'Package architecture "%s" not compatible '
                            'with system architecture "%s"' %
                            (architecture, dpkg_architecture))

            # This isn't ideally quick, since it has to decompress the data
            # part of the package, but dpkg's path filtering code assumes
            # that all paths start with "./" so we must check it before
            # passing the package to dpkg.
            for data_name in package.data:
                if data_name != "." and not data_name.startswith("./"):
                    raise ClickInstallerAuditError(
                        'File name "%s" in package does not start with "./"' %
                        data_name)

            if slow:
                temp_dir = tempfile.mkdtemp(prefix="click")
                try:
                    self.extract(path, temp_dir)
                    command = [
                        "md5sum", "-c", "--quiet",
                        os.path.join("DEBIAN", "md5sums"),
                    ]
                    subprocess.check_call(command, cwd=temp_dir)
                finally:
                    shutil.rmtree(temp_dir)

            return package_name, package_version

    def _drop_privileges(self, username):
        if os.geteuid() != 0:
            return
        pw = pwd.getpwnam(username)
        os.setgroups(
            [g.gr_gid for g in grp.getgrall() if username in g.gr_mem])
        # Portability note: this assumes that we have [gs]etres[gu]id, which
        # is true on Linux but not necessarily elsewhere.  If you need to
        # support something else, there are reasonably standard alternatives
        # involving other similar calls; see e.g. gnulib/lib/idpriv-drop.c.
        os.setresgid(pw.pw_gid, pw.pw_gid, pw.pw_gid)
        os.setresuid(pw.pw_uid, pw.pw_uid, pw.pw_uid)
        assert os.getresuid() == (pw.pw_uid, pw.pw_uid, pw.pw_uid)
        assert os.getresgid() == (pw.pw_gid, pw.pw_gid, pw.pw_gid)
        os.umask(0o022)

    def _euid_access(self, username, path, mode):
        """Like os.access, but for the effective UID."""
        # TODO: Dropping privileges and calling
        # os.access(effective_ids=True) ought to work, but for some reason
        # appears not to return False when it should.  It seems that we need
        # a subprocess to check this reliably.  At least we don't have to
        # exec anything.
        pid = os.fork()
        if pid == 0:  # child
            self._drop_privileges(username)
            os._exit(0 if os.access(path, mode) else 1)
        else:  # parent
            _, status = os.waitpid(pid, 0)
            return status == 0

    def _check_write_permissions(self, path):
        while True:
            if os.path.exists(path):
                break
            path = os.path.dirname(path)
            if path == "/":
                break
        if not self._euid_access("clickpkg", path, os.W_OK):
            raise ClickInstallerPermissionDenied(
                'Cannot acquire permission to write to %s; either run as root '
                'with --user, or use "pkcon install-local" instead' % path)

    def _install_preexec(self, inst_dir):
        self._drop_privileges("clickpkg")

        admin_dir = os.path.join(inst_dir, ".click")
        if not os.path.exists(admin_dir):
            os.makedirs(admin_dir)
            with open(os.path.join(admin_dir, "available"), "w"):
                pass
            with open(os.path.join(admin_dir, "status"), "w"):
                pass
            os.mkdir(os.path.join(admin_dir, "info"))
            os.mkdir(os.path.join(admin_dir, "updates"))
            os.mkdir(os.path.join(admin_dir, "triggers"))

    def _unpack(self, path, user=None, all_users=False, quiet=True):
        package_name, package_version = self.audit(path, check_arch=True)

        # Is this package already unpacked in an underlay (non-topmost)
        # database?
        if self.db.has_package_version(package_name, package_version):
            overlay = self.db.get(self.db.props.size - 1)
            if not overlay.has_package_version(package_name, package_version):
                return package_name, package_version, None

        package_dir = os.path.join(self.db.props.overlay, package_name)
        inst_dir = os.path.join(package_dir, package_version)
        assert (
            os.path.dirname(os.path.dirname(inst_dir)) ==
            self.db.props.overlay)

        self._check_write_permissions(self.db.props.overlay)
        root_click = os.path.join(self.db.props.overlay, ".click")
        if not os.path.exists(root_click):
            os.makedirs(root_click)
            if os.geteuid() == 0:
                pw = pwd.getpwnam("clickpkg")
                os.chown(root_click, pw.pw_uid, pw.pw_gid)

        # TODO: sandbox so that this can only write to the unpack directory
        command = [
            "dpkg",
            # We normally run dpkg as non-root.
            "--force-not-root",
            # /sbin and /usr/sbin may not necessarily be on $PATH; we don't
            # use the tools dpkg gets from there.
            "--force-bad-path",
            # We check the package architecture ourselves in audit().
            "--force-architecture",
            "--instdir", inst_dir,
            "--admindir", os.path.join(inst_dir, ".click"),
            "--path-exclude", "*/.click/*",
            "--log", os.path.join(root_click, "log"),
            "--no-triggers",
            "--install", path,
        ]
        with open(path, "rb") as fd:
            env = dict(os.environ)
            preloads = [self._preload_path()]
            if "LD_PRELOAD" in env:
                preloads.append(env["LD_PRELOAD"])
            env["LD_PRELOAD"] = " ".join(preloads)
            env["CLICK_BASE_DIR"] = self.db.props.overlay
            env["CLICK_PACKAGE_PATH"] = path
            env["CLICK_PACKAGE_FD"] = str(fd.fileno())
            env.pop("HOME", None)
            kwargs = {}
            if sys.version >= "3.2":
                kwargs["pass_fds"] = (fd.fileno(),)
            if quiet:
                fn = subprocess.check_output
                kwargs["stderr"] = subprocess.STDOUT
            else:
                fn = subprocess.check_call
            try:
                fn(command,
                    preexec_fn=partial(self._install_preexec, inst_dir),
                    env=env, universal_newlines=True,
                    **kwargs)
            except subprocess.CalledProcessError as e:
                logging.error("%s failed with exit_code %s:\n%s" % (
                    command, e.returncode, e.output))
                raise
        for dirpath, dirnames, filenames in os.walk(inst_dir):
            for entry in dirnames + filenames:
                entry_path = os.path.join(dirpath, entry)
                entry_mode = os.lstat(entry_path).st_mode
                new_entry_mode = entry_mode | stat.S_IRGRP | stat.S_IROTH
                if entry_mode & stat.S_IXUSR:
                    new_entry_mode |= stat.S_IXGRP | stat.S_IXOTH
                if new_entry_mode != entry_mode:
                    try:
                        os.chmod(entry_path, new_entry_mode)
                    except OSError:
                        pass

        current_path = os.path.join(package_dir, "current")

        if os.path.islink(current_path):
            old_version = os.readlink(current_path)
            if "/" in old_version:
                old_version = None
        else:
            old_version = None
        Click.package_install_hooks(
            self.db, package_name, old_version, package_version,
            user_name=None)

        new_path = os.path.join(package_dir, "current.new")
        Click.symlink_force(package_version, new_path)
        if os.geteuid() == 0:
            # shutil.chown would be more convenient, but it doesn't support
            # follow_symlinks=False in Python 3.3.
            # http://bugs.python.org/issue18108
            pw = pwd.getpwnam("clickpkg")
            os.chown(new_path, pw.pw_uid, pw.pw_gid, follow_symlinks=False)
        os.rename(new_path, current_path)

        return package_name, package_version, old_version

    def install(self, path, user=None, all_users=False, quiet=True):
        package_name, package_version, old_version = self._unpack(
            path, user=user, all_users=all_users, quiet=quiet)

        if user is not None or all_users:
            if all_users:
                registry = Click.User.for_all_users(self.db)
            else:
                registry = Click.User.for_user(self.db, name=user)
            registry.set_version(package_name, package_version)
        else:
            print(dedent("""\
                %s %s has not been registered for any users.
                It may be garbage-collected the next time the system starts.
                To avoid this, use "click register".
                """) % (package_name, package_version))

        if old_version is not None:
            self.db.maybe_remove(package_name, old_version)



# Copyright (C) 2014 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for click_package.framework."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'TestClickFramework',
    ]


import os

from gi.repository import Click

from click_package.tests.helpers import TestCase, touch


class TestClickFramework(TestCase):
    def setUp(self):
        super(TestClickFramework, self).setUp()
        self.use_temp_dir()

    def _setup_frameworks(self, preloads, frameworks_dir=None, frameworks={}):
        if frameworks_dir is None:
            frameworks_dir = os.path.join(self.temp_dir, "frameworks")
        Click.ensuredir(frameworks_dir)
        for framework_name in frameworks:
            framework_path = os.path.join(
                frameworks_dir, "%s.framework" % framework_name)
            with open(framework_path, "w") as framework:
                for key, value in frameworks[framework_name].items():
                    print("%s: %s" % (key, value), file=framework)
        preloads["click_get_frameworks_dir"].side_effect = (
            lambda: self.make_string(frameworks_dir))

    def test_open(self):
        with self.run_in_subprocess(
                "click_get_frameworks_dir") as (enter, preloads):
            enter()
            self._setup_frameworks(preloads, frameworks={"framework-1": {}})
            Click.Framework.open("framework-1")
            self.assertRaisesFrameworkError(
                Click.FrameworkError.NO_SUCH_FRAMEWORK,
                Click.Framework.open, "framework-2")

    def test_has_framework(self):
        with self.run_in_subprocess(
                "click_get_frameworks_dir") as (enter, preloads):
            enter()
            self._setup_frameworks(preloads, frameworks={"framework-1": {}})
            self.assertTrue(Click.Framework.has_framework("framework-1"))
            self.assertFalse(Click.Framework.has_framework("framework-2"))

    def test_get_frameworks(self):
        with self.run_in_subprocess(
                "click_get_frameworks_dir") as (enter, preloads):
            enter()
            self._setup_frameworks(
                preloads,
                frameworks={"ubuntu-sdk-13.10": {}, "ubuntu-sdk-14.04": {},
                            "ubuntu-sdk-14.10": {}})
            self.assertEqual(
                ["ubuntu-sdk-13.10", "ubuntu-sdk-14.04", "ubuntu-sdk-14.10"],
                sorted(f.props.name for f in Click.Framework.get_frameworks()))

    def test_get_frameworks_nonexistent(self):
        with self.run_in_subprocess(
                "click_get_frameworks_dir") as (enter, preloads):
            enter()
            frameworks_dir = os.path.join(self.temp_dir, "nonexistent")
            preloads["click_get_frameworks_dir"].side_effect = (
                lambda: self.make_string(frameworks_dir))
            self.assertEqual([], Click.Framework.get_frameworks())

    def test_get_frameworks_not_directory(self):
        with self.run_in_subprocess(
                "click_get_frameworks_dir") as (enter, preloads):
            enter()
            path = os.path.join(self.temp_dir, "file")
            touch(path)
            preloads["click_get_frameworks_dir"].side_effect = (
                lambda: self.make_string(path))
            self.assertEqual([], Click.Framework.get_frameworks())

    def test_get_frameworks_ignores_other_files(self):
        with self.run_in_subprocess(
                "click_get_frameworks_dir") as (enter, preloads):
            enter()
            frameworks_dir = os.path.join(self.temp_dir, "frameworks")
            Click.ensuredir(frameworks_dir)
            touch(os.path.join(frameworks_dir, "file"))
            preloads["click_get_frameworks_dir"].side_effect = (
                lambda: self.make_string(frameworks_dir))
            self.assertEqual([], Click.Framework.get_frameworks())

    def test_get_frameworks_ignores_unopenable_files(self):
        with self.run_in_subprocess(
                "click_get_frameworks_dir") as (enter, preloads):
            enter()
            frameworks_dir = os.path.join(self.temp_dir, "frameworks")
            Click.ensuredir(frameworks_dir)
            os.symlink(
                "nonexistent", os.path.join(frameworks_dir, "foo.framework"))
            preloads["click_get_frameworks_dir"].side_effect = (
                lambda: self.make_string(frameworks_dir))
            self.assertEqual([], Click.Framework.get_frameworks())

    def test_fields(self):
        with self.run_in_subprocess(
                "click_get_frameworks_dir") as (enter, preloads):
            enter()
            self._setup_frameworks(
                preloads,
                frameworks={
                    "ubuntu-sdk-14.04-qml": {
                        "base-name": "ubuntu-sdk", "base-version": "14.04",
                    }})
            framework = Click.Framework.open("ubuntu-sdk-14.04-qml")
            self.assertCountEqual(
                ["base-name", "base-version"], framework.get_fields())
            self.assertEqual("ubuntu-sdk", framework.get_field("base-name"))
            self.assertEqual("14.04", framework.get_field("base-version"))
            self.assertRaisesFrameworkError(
                Click.FrameworkError.MISSING_FIELD,
                framework.get_field, "nonexistent")
            self.assertEqual("ubuntu-sdk", framework.get_base_name())
            self.assertEqual("14.04", framework.get_base_version())



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Unit tests for click_package.build."""

from __future__ import print_function

__metaclass__ = type
__all__ = [
    'TestClickBuilder',
    'TestClickSourceBuilder',
    ]


import json
import os
import stat
import subprocess
import tarfile
from textwrap import dedent

from click_package.build import ClickBuildError, ClickBuilder, ClickSourceBuilder
from click_package.preinst import static_preinst
from click_package.tests.helpers import (
    disable_logging,
    mkfile,
    TestCase,
    touch,
)


# BAW 2013-04-15: Some tests require umask 022.  Use this decorator to
# temporarily tweak the process's umask.  The test -- or system -- should
# probably be made more robust instead.
def umask(force_umask):
    def decorator(func):
        def wrapper(*args, **kws):
            old_umask = os.umask(force_umask)
            try:
                return func(*args, **kws)
            finally:
                os.umask(old_umask)
        return wrapper
    return decorator


class TestClickBuilderBaseMixin:
    def test_read_manifest(self):
        self.use_temp_dir()
        manifest_path = os.path.join(self.temp_dir, "manifest.json")
        with mkfile(manifest_path) as manifest:
            print(dedent("""\
                {
                    "name": "com.example.test",
                    "version": "1.0",
                    "maintainer": "Foo Bar <foo@example.org>",
                    "title": "test title",
                    "framework": "ubuntu-sdk-13.10"
                }"""), file=manifest)
        self.builder.read_manifest(manifest_path)
        self.assertEqual("com.example.test", self.builder.name)
        self.assertEqual("1.0", self.builder.version)
        self.assertEqual("Foo Bar <foo@example.org>", self.builder.maintainer)
        self.assertEqual("test title", self.builder.title)
        self.assertEqual("all", self.builder.architecture)

    def test_add_file(self):
        self.builder.add_file("/nonexistent", "target")
        self.assertEqual({"/nonexistent": "target"}, self.builder.file_map)

    def test_epochless_version(self):
        self.use_temp_dir()
        manifest_path = os.path.join(self.temp_dir, "manifest.json")
        for version, epochless_version in (
            ("1.0", "1.0"),
            ("1:1.2.3", "1.2.3"),
        ):
            with mkfile(manifest_path) as manifest:
                print(dedent("""\
                    {
                        "name": "com.example.test",
                        "version": "%s",
                        "maintainer": "Foo Bar <foo@example.org>",
                        "title": "test title",
                        "framework": "ubuntu-sdk-13.10"
                    }""") % version, file=manifest)
            self.builder.read_manifest(manifest_path)
            self.assertEqual(epochless_version, self.builder.epochless_version)

    def test_manifest_syntax_error(self):
        self.use_temp_dir()
        manifest_path = os.path.join(self.temp_dir, "manifest.json")
        with mkfile(manifest_path) as manifest:
            # The comma after the "name" entry is intentionally missing.
            print(dedent("""\
                {
                    "name": "com.example.test"
                    "version": "1.0"
                }"""), file=manifest)
        self.assertRaises(
            ClickBuildError, self.builder.read_manifest, manifest_path)


class TestClickBuilder(TestCase, TestClickBuilderBaseMixin):
    def setUp(self):
        super(TestClickBuilder, self).setUp()
        self.builder = ClickBuilder()

    def extract_field(self, path, name):
        return subprocess.check_output(
            ["dpkg-deb", "-f", path, name],
            universal_newlines=True).rstrip("\n")

    @disable_logging
    @umask(0o22)
    def test_build(self):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        with mkfile(os.path.join(scratch, "bin", "foo")) as f:
            f.write("test /bin/foo\n")
        os.symlink("foo", os.path.join(scratch, "bin", "bar"))
        touch(os.path.join(scratch, ".git", "config"))
        with mkfile(os.path.join(scratch, "toplevel")) as f:
            f.write("test /toplevel\n")
        os.symlink(
            "file-does-not-exist", os.path.join(scratch, "broken-symlink"))
        with mkfile(os.path.join(scratch, "manifest.json")) as f:
            json.dump({
                "name": "com.example.test",
                "version": "1.0",
                "maintainer": "Foo Bar <foo@example.org>",
                "title": "test title",
                "architecture": "all",
                "framework": "ubuntu-sdk-13.10",
            }, f)
            # build() overrides this back to 0o644
            os.fchmod(f.fileno(), 0o600)
        self.builder.add_file(scratch, "/")
        path = os.path.join(self.temp_dir, "com.example.test_1.0_all.click")
        self.assertEqual(path, self.builder.build(self.temp_dir))
        self.assertTrue(os.path.exists(path))
        for key, value in (
            ("Package", "com.example.test"),
            ("Version", "1.0"),
            ("Click-Version", "0.4"),
            ("Architecture", "all"),
            ("Maintainer", "Foo Bar <foo@example.org>"),
            ("Description", "test title"),
        ):
            self.assertEqual(value, self.extract_field(path, key))
        self.assertNotEqual(
            "", self.extract_field(path, "Installed-Size"))
        control_path = os.path.join(self.temp_dir, "control")
        subprocess.check_call(["dpkg-deb", "-e", path, control_path])
        manifest_path = os.path.join(control_path, "manifest")
        self.assertEqual(0o644, stat.S_IMODE(os.stat(manifest_path).st_mode))
        with open(os.path.join(scratch, "manifest.json")) as source, \
                open(manifest_path) as target:
            source_json = json.load(source)
            target_json = json.load(target)
            self.assertNotEqual("", target_json["installed-size"])
            del target_json["installed-size"]
            self.assertEqual(source_json, target_json)
        with open(os.path.join(control_path, "md5sums")) as md5sums:
            self.assertRegex(
                md5sums.read(),
                r"^"
                r"eb774c3ead632b397d6450d1df25e001  bin/bar\n"
                r"eb774c3ead632b397d6450d1df25e001  bin/foo\n"
                r"49327ce6306df8a87522456b14a179e0  toplevel\n"
                r"$")
        with open(os.path.join(control_path, "preinst")) as preinst:
            self.assertEqual(static_preinst, preinst.read())
        contents = subprocess.check_output(
            ["dpkg-deb", "-c", path], universal_newlines=True)
        self.assertRegex(contents, r"^drwxr-xr-x root/root         0 .* \./\n")
        self.assertRegex(
            contents,
            "\nlrwxrwxrwx root/root         0 .* \./bin/bar -> foo\n")
        self.assertRegex(
            contents, "\n-rw-r--r-- root/root        14 .* \./bin/foo\n")
        self.assertRegex(
            contents, "\n-rw-r--r-- root/root        15 .* \./toplevel\n")
        extract_path = os.path.join(self.temp_dir, "extract")
        subprocess.check_call(["dpkg-deb", "-x", path, extract_path])
        for rel_path in (
            os.path.join("bin", "foo"),
            "toplevel",
        ):
            with open(os.path.join(scratch, rel_path)) as source, \
                    open(os.path.join(extract_path, rel_path)) as target:
                self.assertEqual(source.read(), target.read())
        self.assertTrue(
            os.path.islink(os.path.join(extract_path, "bin", "bar")))
        self.assertEqual(
            "foo", os.readlink(os.path.join(extract_path, "bin", "bar")))

    def _make_scratch_dir(self, manifest_override={}):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        manifest = {
                "name": "com.example.test",
                "version": "1.0",
                "maintainer": "Foo Bar <foo@example.org>",
                "title": "test title",
                "architecture": "all",
                "framework": "ubuntu-sdk-13.10",
        }
        manifest.update(manifest_override)
        with mkfile(os.path.join(scratch, "manifest.json")) as f:
            json.dump(manifest, f)
        self.builder.add_file(scratch, "/")
        return scratch

    @disable_logging
    def test_build_excludes_dot_click(self):
        scratch = self._make_scratch_dir()
        touch(os.path.join(scratch, ".click", "evil-file"))
        path = self.builder.build(self.temp_dir)
        extract_path = os.path.join(self.temp_dir, "extract")
        subprocess.check_call(["dpkg-deb", "-x", path, extract_path])
        self.assertEqual([], os.listdir(extract_path))

    def test_build_ignore_pattern(self):
        scratch = self._make_scratch_dir()
        touch(os.path.join(scratch, "build", "foo.o"))
        self.builder.add_file(scratch, "/")
        self.builder.add_ignore_pattern("build")
        path = self.builder.build(self.temp_dir)
        extract_path = os.path.join(self.temp_dir, "extract")
        subprocess.check_call(["dpkg-deb", "-x", path, extract_path])
        self.assertEqual([], os.listdir(extract_path))

    @disable_logging
    def test_build_multiple_architectures(self):
        scratch = self._make_scratch_dir(manifest_override={
                "architecture": ["armhf", "i386"],
        })
        path = os.path.join(self.temp_dir, "com.example.test_1.0_multi.click")
        self.assertEqual(path, self.builder.build(self.temp_dir))
        self.assertTrue(os.path.exists(path))
        self.assertEqual("multi", self.extract_field(path, "Architecture"))
        control_path = os.path.join(self.temp_dir, "control")
        subprocess.check_call(["dpkg-deb", "-e", path, control_path])
        manifest_path = os.path.join(control_path, "manifest")
        with open(os.path.join(scratch, "manifest.json")) as source, \
                open(manifest_path) as target:
            source_json = json.load(source)
            target_json = json.load(target)
            del target_json["installed-size"]
            self.assertEqual(source_json, target_json)

    @disable_logging
    def test_build_multiple_frameworks(self):
        scratch = self._make_scratch_dir(manifest_override={
                "framework":
                    "ubuntu-sdk-14.04-basic, ubuntu-sdk-14.04-webapps",
        })
        path = self.builder.build(self.temp_dir)
        control_path = os.path.join(self.temp_dir, "control")
        subprocess.check_call(["dpkg-deb", "-e", path, control_path])
        manifest_path = os.path.join(control_path, "manifest")
        with open(os.path.join(scratch, "manifest.json")) as source, \
                open(manifest_path) as target:
            source_json = json.load(source)
            target_json = json.load(target)
            del target_json["installed-size"]
            self.assertEqual(source_json, target_json)


class TestClickFrameworkValidation(TestCase):
    def setUp(self):
        super(TestClickFrameworkValidation, self).setUp()
        self.builder = ClickBuilder()
        for framework_name in ("ubuntu-sdk-13.10",
                               "ubuntu-sdk-14.04-papi",
                               "ubuntu-sdk-14.04-html",
                               "docker-sdk-1.3"):
            self._create_mock_framework_file(framework_name)

    def test_validate_framework_good(self):
        valid_framework_values = (
            "ubuntu-sdk-13.10",
            "ubuntu-sdk-14.04-papi, ubuntu-sdk-14.04-html",
            "ubuntu-sdk-13.10, docker-sdk-1.3",
        )
        for framework in valid_framework_values:
            self.builder._validate_framework(framework)

    def test_validate_framework_bad(self):
        invalid_framework_values = (
            "ubuntu-sdk-13.10, ubuntu-sdk-14.04-papi",
            "ubuntu-sdk-13.10 (>= 13.10)",
            "ubuntu-sdk-13.10 | ubuntu-sdk-14.04",
        )
        for framework in invalid_framework_values:
            with self.assertRaises(ClickBuildError):
                self.builder._validate_framework(framework)


class TestClickSourceBuilder(TestCase, TestClickBuilderBaseMixin):
    def setUp(self):
        super(TestClickSourceBuilder, self).setUp()
        self.builder = ClickSourceBuilder()

    @umask(0o22)
    def test_build(self):
        self.use_temp_dir()
        scratch = os.path.join(self.temp_dir, "scratch")
        touch(os.path.join(scratch, "bin", "foo"))
        touch(os.path.join(scratch, ".git", "config"))
        touch(os.path.join(scratch, "foo.so"))
        touch(os.path.join(scratch, "build", "meep.goah"))
        with mkfile(os.path.join(scratch, "manifest.json")) as f:
            json.dump({
                "name": "com.example.test",
                "version": "1.0",
                "maintainer": "Foo Bar <foo@example.org>",
                "title": "test title",
                "architecture": "all",
                "framework": "ubuntu-sdk-13.10",
            }, f)
            # build() overrides this back to 0o644
            os.fchmod(f.fileno(), 0o600)
        self.builder.add_file(scratch, "./")
        self.builder.add_ignore_pattern("build")
        path = os.path.join(self.temp_dir, "com.example.test_1.0.tar.gz")
        self.assertEqual(path, self.builder.build(self.temp_dir))
        self.assertTrue(os.path.exists(path))
        with tarfile.open(path, mode="r:gz") as tar:
            self.assertCountEqual(
                [".", "./bin", "./bin/foo", "./manifest.json"], tar.getnames())
            self.assertTrue(tar.getmember("./bin/foo").isfile())



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for the click chroot feature."""

from glob import glob
import json
import os
import shutil
import subprocess

from six import with_metaclass

from .helpers import (
    chdir,
    require_network,
    require_overlay,
    require_root,
    ClickTestCase,
)

# the branches we want to testbuild
TEST_BUILD_BRANCHES = [
    "lp:camera-app",
    "lp:notes-app",
]

# command to "configure"
CORE_APP_CONFIGURE_CMD = [
    "cmake", "..", "-DCLICK_MODE=on", "-DINSTALL_TESTS=off"]

# command to make install
CLICK_TARGET_DIR = "click-package"
CORE_APP_MAKE_CMD = [
    "make", "DESTDIR=%s" % CLICK_TARGET_DIR, "install"]

# architectures with native- or cross-compiling support for armhf
ALLOW_ARCHITECTURES = ["amd64", "arm64", "armhf", "i386"]


def find_manifest(start_dir):
    """Find a click manifest.json{,.in} under the given directory"""
    for path, dirs, files in os.walk(start_dir):
        for needle in ["manifest.json", "manifest.json.in"]:
            if needle in files:
                return os.path.join(path, needle)
    return None


class AddBranchTestFunctions(type):
    """Metaclass that creates one test for each branch"""
    def __new__(cls, name, bases, dct):
        for branch in TEST_BUILD_BRANCHES:
            name = "test_build_%s" % branch.split(":")[1].replace("-", "_")
            dct[name] = lambda self: self._testbuild_branch(branch)
        return type.__new__(cls, name, bases, dct)


class TestBuildCoreApps(with_metaclass(AddBranchTestFunctions, ClickTestCase)):

    @classmethod
    def setUpClass(cls):
        super(TestBuildCoreApps, cls).setUpClass()
        require_root()
        require_network()
        require_overlay()

    def _run_in_chroot(self, cmd):
        """Run the given cmd in a click chroot"""
        return subprocess.check_call(self.chroot_cmd + ["run"] + cmd)

    def _set_arch_and_framework_from_manifest(self, manifest):
        with open(manifest) as f:
            data = json.load(f)
        self.arch = data["architecture"]
        self.framework = data["framework"]

    @property
    def chroot_cmd(self):
        return [
            self.click_binary, "chroot", "-a", self.arch, "-f", self.framework]

    def _ensure_click_chroot(self):
        if subprocess.call(self.chroot_cmd + ["exists"]) != 0:
            subprocess.check_call(self.chroot_cmd + ["create"])

    def configure(self):
        self._run_in_chroot(CORE_APP_CONFIGURE_CMD)

    def make(self):
        self._run_in_chroot(CORE_APP_MAKE_CMD)

    def create_click(self):
        subprocess.check_call(
            [self.click_binary, "build", CLICK_TARGET_DIR])
        # we expect exactly one click
        self.assertEqual(len(glob("*.click")), 1)

    def _testbuild_branch(self, branch):
        system_arch = subprocess.check_output(
            ["dpkg", "--print-architecture"], universal_newlines=True).strip()
        if system_arch not in ALLOW_ARCHITECTURES:
            self.skipTest("%s has no armhf build support" % system_arch)
        # get and parse
        branch_dir = branch[len("lp:"):]
        build_dir = os.path.join(branch_dir, "build-tree")
        if os.path.exists(branch_dir):
            subprocess.check_call(["bzr", "pull"], cwd=branch_dir)
        else:
            subprocess.check_call(["bzr", "branch", branch])
        manifest = find_manifest(branch_dir)
        # build it
        self._set_arch_and_framework_from_manifest(manifest)
        if os.path.exists(build_dir):
            shutil.rmtree(build_dir)
        os.makedirs(build_dir)
        with chdir(build_dir):
            self._ensure_click_chroot()
            self.configure()
            self.make()
            self.create_click()



# Copyright (C) 2014 Canonical Ltd.
# Author: Michael Vogt <michael.vogt@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Integration tests for the click CLI list command."""

import os
import subprocess

from .helpers import ClickTestCase


class TestList(ClickTestCase):
    def test_list_simple(self):
        name = "com.ubuntu.verify-ok"
        path_to_click = self._make_click(name, framework="")
        user = os.environ.get("USER", "root")
        self.click_install(path_to_click, name, user)
        output = subprocess.check_output(
            [self.click_binary, "list", "--user=%s" % user],
            universal_newlines=True)
        self.assertIn(name, output)



from __future__ import print_function

import os
import sys

from click_package.tests import config


def _append_env_path(envname, value):
    if envname in os.environ:
        if value in os.environ[envname].split(":"):
            return False
        os.environ[envname] = "%s:%s" % (os.environ[envname], value)
    else:
        os.environ[envname] = value
    return True


def get_executable():
    """Get python executable (respecting if python-coverage was used)"""
    coverage_executable = sys.executable+"-coverage"
    if "coverage" in sys.modules and os.path.isfile(coverage_executable):
        return [coverage_executable, "run", "-p"]
    return [sys.executable]


# Don't do any of this in interactive mode.
if not hasattr(sys, "ps1"):
    _lib_click_dir = os.path.join(config.abs_top_builddir, "lib", "click")
    changed = False
    if _append_env_path(
            "LD_LIBRARY_PATH", os.path.join(_lib_click_dir, ".libs")):
        changed = True
    if _append_env_path("GI_TYPELIB_PATH", _lib_click_dir):
        changed = True
    if changed:
        coverage_executable = get_executable()
        # We have to re-exec ourselves to get the dynamic loader to pick up
        # the new value of LD_LIBRARY_PATH.
        if "-m unittest" in sys.argv[0]:
            # unittest does horrible things to sys.argv in the name of
            # "usefulness", making the re-exec more painful than it needs to
            # be.
            os.execvp(
                coverage_executable[0], coverage_executable + ["-m", "unittest"] + sys.argv[1:])
        else:
            os.execvp(coverage_executable[0], coverage_executable + sys.argv)
        os._exit(1)



# Copyright (C) 2013 Canonical Ltd.
# Author: Colin Watson <cjwatson@ubuntu.com>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; version 3 of the License.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Install or remove a Click system hook."""

from __future__ import print_function

from optparse import OptionParser
import sys
from textwrap import dedent

from gi.repository import Click, GLib


per_hook_subcommands = {
    "install": "install",
    "remove": "remove",
    }


def run(argv):
    parser = OptionParser(dedent("""\
        %prog hook [options] SUBCOMMAND [...]

        Subcommands are as follows:

          install HOOK
          remove HOOK
          run-system
          run-user [--user=USER]"""))
    parser.add_option(
        "--root", metavar="PATH", help="look for additional packages in PATH")
    parser.add_option(
        "--user", metavar="USER",
        help=(
            "run user-level hooks for USER (default: current user; only "
            "applicable to run-user)"))
    options, args = parser.parse_args(argv)
    if len(args) < 1:
        parser.error("need subcommand (install, remove, run-system, run-user)")
    subcommand = args[0]
    if subcommand in per_hook_subcommands:
        if len(args) < 2:
            parser.error("need hook name")
        db = Click.DB()
        db.read(db_dir=None)
        if options.root is not None:
            db.add(options.root)
        name = args[1]
        hook = Click.Hook.open(db, name)
        getattr(hook, per_hook_subcommands[subcommand])(user_name=None)
    elif subcommand == "run-system":
        db = Click.DB()
        db.read(db_dir=None)
        if options.root is not None:
            db.add(options.root)
        try:
            Click.run_system_hooks(db)
        except GLib.GError as e:
            if e.domain == "click_hooks_error-quark":
                print(e.message, file=sys.stderr)
                return 1
            else:
                raise
    elif subcommand == "run-user":
        db = Click.DB()
        db.read(db_dir=None)
        if options.root is not None:
            db.add(options.root)
        try:
            Click.run_user_hooks(db, user_name=options.user)
        except GLib.GError as e:
            if e.domain == "click_hooks_error-quark":
                print(e.message, file=sys.stderr)
                return 1
            else:
                raise
    else:
        parser.error(
            "unknown subcommand '%s' (known: install, remove, run-system,"
            "run-user)" % subcommand)
    return 0
